"""
Test case for the consul execution module
"""

import logging

import pytest

import salt.modules.consul as consul
import salt.utils.http
import salt.utils.json
import salt.utils.platform
from salt.exceptions import SaltInvocationError
from tests.support.mock import MagicMock, patch

log = logging.getLogger(__name__)


@pytest.fixture
def configure_loader_modules():
    return {
        consul: {
            "__opts__": {"consul": {"url": "http://127.0.0.1", "token": "test_token"}},
            "__grains__": {"id": "test-minion"},
        }
    }


def test_list():
    """
    Test salt.modules.consul.list function
    """
    mock_query = MagicMock(return_value={"data": ["foo"], "res": True})
    with patch.object(consul, "_query", mock_query):
        consul_return = consul.list_(consul_url="http://127.0.0.1", token="test_token")

    assert consul_return == {"data": ["foo"], "res": True}


def test_get():
    """
    Test salt.modules.consul.get function
    """
    #
    # No key argument results in SaltInvocationError, exception
    #
    with pytest.raises(SaltInvocationError):
        consul.put(consul_url="http://127.0.0.1", token="test_token")

    mock_query = MagicMock(
        return_value={
            "data": [
                {
                    "LockIndex": 0,
                    "Key": "foo",
                    "Flags": 0,
                    "Value": "YmFy",
                    "CreateIndex": 128,
                    "ModifyIndex": 128,
                },
            ],
            "res": True,
        }
    )
    with patch.object(consul, "_query", mock_query):
        consul_return = consul.get(
            consul_url="http://127.0.0.1", key="foo", token="test_token"
        )
    _expected = {
        "data": [
            {
                "CreateIndex": 128,
                "Flags": 0,
                "Key": "foo",
                "LockIndex": 0,
                "ModifyIndex": 128,
                "Value": "YmFy",
            }
        ],
        "res": True,
    }

    assert consul_return == _expected

    mock_query = MagicMock(
        return_value={
            "data": [
                {
                    "LockIndex": 0,
                    "Key": "foo",
                    "Flags": 0,
                    "Value": "b'bar'",
                    "CreateIndex": 128,
                    "ModifyIndex": 128,
                },
            ],
            "res": True,
        }
    )
    with patch.object(consul, "_query", mock_query):
        consul_return = consul.get(
            consul_url="http://127.0.0.1", key="foo", token="test_token"
        )
    _expected = {
        "data": [
            {
                "CreateIndex": 128,
                "Flags": 0,
                "Key": "foo",
                "LockIndex": 0,
                "ModifyIndex": 128,
                "Value": "b'bar'",
            }
        ],
        "res": True,
    }

    assert consul_return == _expected


def test_put():
    """
    Test salt.modules.consul.put function
    """
    #
    # No key argument results in SaltInvocationError, exception
    #
    with pytest.raises(SaltInvocationError):
        consul.put(consul_url="http://127.0.0.1", token="test_token")

    #
    # Test when we're unable to connect to Consul
    #
    mock_consul_get = {
        "data": [
            {
                "LockIndex": 0,
                "Key": "web/key1",
                "Flags": 0,
                "Value": "ImhlbGxvIHRoZXJlIg==",
                "CreateIndex": 299,
                "ModifyIndex": 299,
            }
        ],
        "res": True,
    }
    with patch.object(consul, "session_list", MagicMock(return_value=[])):
        with patch.object(consul, "get", MagicMock(return_value=mock_consul_get)):
            ret = consul.put(
                consul_url="http://127.0.0.1:8501",
                token="test_token",
                key="web/key1",
                value="Hello world",
            )
    expected_res = (False,)
    expected_data = "Unable to add key web/key1 with value Hello world."
    if salt.utils.platform.is_windows():
        expected_error = "Unknown error"
    else:
        expected_error = "Connection refused"
    assert not ret["res"]
    assert expected_data == ret["data"]
    assert expected_error in ret["error"]

    #
    # Working as expected
    #
    mock_query = MagicMock(
        return_value={
            "data": [
                {
                    "LockIndex": 0,
                    "Key": "foo",
                    "Flags": 0,
                    "Value": "YmFy",
                    "CreateIndex": 128,
                    "ModifyIndex": 128,
                },
            ],
            "res": True,
        }
    )
    with patch.object(consul, "session_list", MagicMock(return_value=[])):
        with patch.object(consul, "get", MagicMock(return_value=mock_consul_get)):
            with patch.object(consul, "_query", mock_query):
                ret = consul.put(
                    consul_url="http://127.0.0.1:8500",
                    token="test_token",
                    key="web/key1",
                    value="Hello world",
                )
    _expected = {"res": True, "data": "Added key web/key1 with value Hello world."}
    assert ret == _expected


def test_delete():
    """
    Test salt.modules.consul.delete function
    """
    #
    # No key argument results in SaltInvocationError, exception
    #
    with pytest.raises(SaltInvocationError):
        consul.put(consul_url="http://127.0.0.1", token="test_token")

    #
    # Test when we're unable to connect to Consul
    #
    ret = consul.delete(
        consul_url="http://127.0.0.1:8501",
        token="test_token",
        key="web/key1",
        value="Hello world",
    )
    expected_res = (False,)
    expected_data = "Unable to delete key web/key1."
    if salt.utils.platform.is_windows():
        expected_error = "Unknown error"
    else:
        expected_error = "Connection refused"
    assert not ret["res"]
    assert expected_data == ret["message"]
    assert expected_error in ret["error"]

    #
    # Working as expected
    #
    mock_query = MagicMock(return_value={"data": True, "res": True})
    with patch.object(consul, "_query", mock_query):
        ret = consul.delete(
            consul_url="http://127.0.0.1:8500",
            token="test_token",
            key="web/key1",
            value="Hello world",
        )
    _expected = {"res": True, "message": "Deleted key web/key1."}
    assert ret == _expected


def test_agent_maintenance():
    """
    Test consul agent maintenance
    """
    consul_url = "http://localhost:1313"
    key = "cluster/key"

    mock_result = "test"
    mock_http_result = {"status": 200, "dict": mock_result}
    mock_http_result_false = {"status": 204, "dict": mock_result}
    mock_url = MagicMock(return_value=consul_url)
    mock_nourl = MagicMock(return_value=None)

    # no consul url error
    with patch.dict(consul.__salt__, {"config.get": mock_nourl}):
        with patch.object(
            salt.modules.consul, "session_list", return_value=mock_result
        ):
            result = consul.agent_maintenance(consul_url="")
            expected = {"message": "No Consul URL found.", "res": False}
            assert expected == result

    # no required argument
    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            with patch.object(
                salt.modules.consul, "session_list", return_value=mock_result
            ):
                msg = 'Required parameter "enable" is missing.'
                result = consul.agent_maintenance(consul_url=consul_url)
                expected = {"message": msg, "res": False}
                assert expected == result

    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            with patch.object(
                salt.modules.consul, "session_list", return_value=mock_result
            ):
                msg = "Agent maintenance mode {}ed."
                value = "enabl"
                result = consul.agent_maintenance(consul_url=consul_url, enable=value)
                expected = {"message": msg.format(value), "res": True}
                assert expected == result

    with patch.object(salt.utils.http, "query", return_value=mock_http_result_false):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            with patch.object(
                salt.modules.consul, "session_list", return_value=mock_result
            ):
                msg = "Unable to change maintenance mode for agent."
                value = "enabl"
                result = consul.agent_maintenance(consul_url=consul_url, enable=value)
                expected = {"message": msg, "res": True}
                assert expected == result


def test_agent_join():
    """
    Test consul agent join
    """
    consul_url = "http://localhost:1313"
    key = "cluster/key"

    mock_result = "test"
    mock_http_result = {"status": 200, "dict": mock_result}
    mock_http_result_false = {"status": 204, "dict": mock_result}
    mock_url = MagicMock(return_value=consul_url)
    mock_nourl = MagicMock(return_value=None)

    # no consul url error
    with patch.dict(consul.__salt__, {"config.get": mock_nourl}):
        with patch.object(
            salt.modules.consul, "session_list", return_value=mock_result
        ):
            result = consul.agent_join(consul_url="")
            expected = {"message": "No Consul URL found.", "res": False}
            assert expected == result

    # no required argument
    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            with patch.object(
                salt.modules.consul, "session_list", return_value=mock_result
            ):
                msg = 'Required parameter "address" is missing.'
                pytest.raises(
                    SaltInvocationError, consul.agent_join, consul_url=consul_url
                )

    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            with patch.object(
                salt.modules.consul, "session_list", return_value=mock_result
            ):
                msg = "Agent joined the cluster"
                result = consul.agent_join(consul_url=consul_url, address="test")
                expected = {"message": msg, "res": True}
                assert expected == result

    with patch.object(salt.utils.http, "query", return_value=mock_http_result_false):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            with patch.object(
                salt.modules.consul, "session_list", return_value=mock_result
            ):
                msg = "Unable to join the cluster."
                value = "enabl"
                result = consul.agent_join(consul_url=consul_url, address="test")
                expected = {"message": msg, "res": False}
                assert expected == result


def test_agent_leave():
    """
    Test consul agent leave
    """
    consul_url = "http://localhost:1313"
    key = "cluster/key"

    mock_result = "test"
    mock_http_result = {"status": 200, "dict": mock_result}
    mock_http_result_false = {"status": 204, "dict": mock_result}
    mock_url = MagicMock(return_value=consul_url)
    mock_nourl = MagicMock(return_value=None)

    # no consul url error
    with patch.dict(consul.__salt__, {"config.get": mock_nourl}):
        with patch.object(
            salt.modules.consul, "session_list", return_value=mock_result
        ):
            result = consul.agent_join(consul_url="")
            expected = {"message": "No Consul URL found.", "res": False}
            assert expected == result

    node = "node1"

    # no required argument
    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            with patch.object(
                salt.modules.consul, "session_list", return_value=mock_result
            ):
                pytest.raises(
                    SaltInvocationError, consul.agent_leave, consul_url=consul_url
                )

    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            with patch.object(
                salt.modules.consul, "session_list", return_value=mock_result
            ):
                msg = "Node {} put in leave state."
                result = consul.agent_leave(consul_url=consul_url, node=node)
                expected = {"message": msg.format(node), "res": True}
                assert expected == result

    with patch.object(salt.utils.http, "query", return_value=mock_http_result_false):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            with patch.object(
                salt.modules.consul, "session_list", return_value=mock_result
            ):
                msg = "Unable to change state for {}."
                result = consul.agent_leave(consul_url=consul_url, node=node)
                expected = {"message": msg.format(node), "res": False}
                assert expected == result


def test_agent_check_register():
    """
    Test consul agent check register
    """
    consul_url = "http://localhost:1313"
    key = "cluster/key"

    mock_result = "test"
    mock_http_result = {"status": 200, "dict": mock_result}
    mock_http_result_false = {"status": 204, "dict": mock_result}
    mock_url = MagicMock(return_value=consul_url)
    mock_nourl = MagicMock(return_value=None)

    # no consul url error
    with patch.dict(consul.__salt__, {"config.get": mock_nourl}):
        with patch.object(
            salt.modules.consul, "session_list", return_value=mock_result
        ):
            result = consul.agent_check_register(consul_url="")
            expected = {"message": "No Consul URL found.", "res": False}
            assert expected == result

    name = "name1"

    # no required arguments
    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            with patch.object(
                salt.modules.consul, "session_list", return_value=mock_result
            ):
                pytest.raises(
                    SaltInvocationError,
                    consul.agent_check_register,
                    consul_url=consul_url,
                )

                # missing script, or http
                msg = 'Required parameter "script" or "http" is missing.'
                result = consul.agent_check_register(consul_url=consul_url, name=name)
                expected = {"message": msg, "res": False}
                assert expected == result

                # missing interval
                msg = 'Required parameter "interval" is missing.'
                result = consul.agent_check_register(
                    consul_url=consul_url,
                    name=name,
                    script="test",
                    http="test",
                    ttl="test",
                )
                expected = {"message": msg, "res": False}
                assert expected == result

    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            with patch.object(
                salt.modules.consul, "session_list", return_value=mock_result
            ):
                msg = "Check {} added to agent."
                result = consul.agent_check_register(
                    consul_url=consul_url,
                    name=name,
                    script="test",
                    http="test",
                    ttl="test",
                    interval="test",
                )
                expected = {"message": msg.format(name), "res": True}
                assert expected == result

    with patch.object(salt.utils.http, "query", return_value=mock_http_result_false):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            with patch.object(
                salt.modules.consul, "session_list", return_value=mock_result
            ):
                msg = "Unable to add check to agent."
                result = consul.agent_check_register(
                    consul_url=consul_url,
                    name=name,
                    script="test",
                    http="test",
                    ttl="test",
                    interval="test",
                )
                expected = {"message": msg.format(name), "res": False}
                assert expected == result


def test_agent_check_deregister():
    """
    Test consul agent check register
    """
    consul_url = "http://localhost:1313"
    key = "cluster/key"

    mock_result = "test"
    mock_http_result = {"status": 200, "dict": mock_result}
    mock_http_result_false = {"status": 204, "dict": mock_result}
    mock_url = MagicMock(return_value=consul_url)
    mock_nourl = MagicMock(return_value=None)

    # no consul url error
    with patch.dict(consul.__salt__, {"config.get": mock_nourl}):
        with patch.object(
            salt.modules.consul, "session_list", return_value=mock_result
        ):
            result = consul.agent_check_register(consul_url="")
            expected = {"message": "No Consul URL found.", "res": False}
            assert expected == result

    checkid = "id1"

    # no required arguments
    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            with patch.object(
                salt.modules.consul, "session_list", return_value=mock_result
            ):
                pytest.raises(
                    SaltInvocationError,
                    consul.agent_check_deregister,
                    consul_url=consul_url,
                )

    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            with patch.object(
                salt.modules.consul, "session_list", return_value=mock_result
            ):
                msg = "Check {} removed from agent."
                result = consul.agent_check_deregister(
                    consul_url=consul_url, checkid=checkid
                )
                expected = {"message": msg.format(checkid), "res": True}
                assert expected == result

    with patch.object(salt.utils.http, "query", return_value=mock_http_result_false):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            with patch.object(
                salt.modules.consul, "session_list", return_value=mock_result
            ):
                msg = "Unable to remove check from agent."
                result = consul.agent_check_deregister(
                    consul_url=consul_url, checkid=checkid
                )
                expected = {"message": msg.format(checkid), "res": False}
                assert expected == result


def test_agent_check_pass():
    """
    Test consul agent check pass
    """
    consul_url = "http://localhost:1313"
    key = "cluster/key"

    mock_result = "test"
    mock_http_result = {"status": 200, "dict": mock_result}
    mock_http_result_false = {"status": 204, "dict": mock_result}
    mock_url = MagicMock(return_value=consul_url)
    mock_nourl = MagicMock(return_value=None)

    # no consul url error
    with patch.dict(consul.__salt__, {"config.get": mock_nourl}):
        with patch.object(
            salt.modules.consul, "session_list", return_value=mock_result
        ):
            result = consul.agent_check_register(consul_url="")
            expected = {"message": "No Consul URL found.", "res": False}
            assert expected == result

    checkid = "id1"

    # no required arguments
    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            with patch.object(
                salt.modules.consul, "session_list", return_value=mock_result
            ):
                pytest.raises(
                    SaltInvocationError,
                    consul.agent_check_pass,
                    consul_url=consul_url,
                )

    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            with patch.object(
                salt.modules.consul, "session_list", return_value=mock_result
            ):
                msg = "Check {} marked as passing."
                result = consul.agent_check_pass(consul_url=consul_url, checkid=checkid)
                expected = {"message": msg.format(checkid), "res": True}
                assert expected == result

    with patch.object(salt.utils.http, "query", return_value=mock_http_result_false):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            with patch.object(
                salt.modules.consul, "session_list", return_value=mock_result
            ):
                msg = "Unable to update check {}."
                result = consul.agent_check_pass(consul_url=consul_url, checkid=checkid)
                expected = {"message": msg.format(checkid), "res": False}
                assert expected == result


def test_agent_check_warn():
    """
    Test consul agent check warn
    """
    consul_url = "http://localhost:1313"
    key = "cluster/key"

    mock_result = "test"
    mock_http_result = {"status": 200, "dict": mock_result}
    mock_http_result_false = {"status": 204, "dict": mock_result}
    mock_url = MagicMock(return_value=consul_url)
    mock_nourl = MagicMock(return_value=None)

    # no consul url error
    with patch.dict(consul.__salt__, {"config.get": mock_nourl}):
        with patch.object(
            salt.modules.consul, "session_list", return_value=mock_result
        ):
            result = consul.agent_check_register(consul_url="")
            expected = {"message": "No Consul URL found.", "res": False}
            assert expected == result

    checkid = "id1"

    # no required arguments
    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            with patch.object(
                salt.modules.consul, "session_list", return_value=mock_result
            ):
                pytest.raises(
                    SaltInvocationError,
                    consul.agent_check_warn,
                    consul_url=consul_url,
                )

    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            with patch.object(
                salt.modules.consul, "session_list", return_value=mock_result
            ):
                msg = "Check {} marked as warning."
                result = consul.agent_check_warn(consul_url=consul_url, checkid=checkid)
                expected = {"message": msg.format(checkid), "res": True}
                assert expected == result

    with patch.object(salt.utils.http, "query", return_value=mock_http_result_false):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            with patch.object(
                salt.modules.consul, "session_list", return_value=mock_result
            ):
                msg = "Unable to update check {}."
                result = consul.agent_check_warn(consul_url=consul_url, checkid=checkid)
                expected = {"message": msg.format(checkid), "res": False}
                assert expected == result


def test_agent_check_fail():
    """
    Test consul agent check warn
    """
    consul_url = "http://localhost:1313"
    key = "cluster/key"

    mock_result = "test"
    mock_http_result = {"status": 200, "dict": mock_result}
    mock_http_result_false = {"status": 204, "dict": mock_result}
    mock_url = MagicMock(return_value=consul_url)
    mock_nourl = MagicMock(return_value=None)

    # no consul url error
    with patch.dict(consul.__salt__, {"config.get": mock_nourl}):
        with patch.object(
            salt.modules.consul, "session_list", return_value=mock_result
        ):
            result = consul.agent_check_register(consul_url="")
            expected = {"message": "No Consul URL found.", "res": False}
            assert expected == result

    checkid = "id1"

    # no required arguments
    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            with patch.object(
                salt.modules.consul, "session_list", return_value=mock_result
            ):
                pytest.raises(
                    SaltInvocationError,
                    consul.agent_check_fail,
                    consul_url=consul_url,
                )

    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            with patch.object(
                salt.modules.consul, "session_list", return_value=mock_result
            ):
                msg = "Check {} marked as critical."
                result = consul.agent_check_fail(consul_url=consul_url, checkid=checkid)
                expected = {"message": msg.format(checkid), "res": True}
                assert expected == result

    with patch.object(salt.utils.http, "query", return_value=mock_http_result_false):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            with patch.object(
                salt.modules.consul, "session_list", return_value=mock_result
            ):
                msg = "Unable to update check {}."
                result = consul.agent_check_fail(consul_url=consul_url, checkid=checkid)
                expected = {"message": msg.format(checkid), "res": False}
                assert expected == result


def test_agent_service_register():
    """
    Test consul agent service register
    """
    consul_url = "http://localhost:1313"
    key = "cluster/key"

    mock_result = "test"
    mock_http_result = {"status": 200, "dict": mock_result}
    mock_http_result_false = {"status": 204, "dict": mock_result}
    mock_url = MagicMock(return_value=consul_url)
    mock_nourl = MagicMock(return_value=None)

    # no consul url error
    with patch.dict(consul.__salt__, {"config.get": mock_nourl}):
        with patch.object(
            salt.modules.consul, "session_list", return_value=mock_result
        ):
            result = consul.agent_service_register(consul_url="")
            expected = {"message": "No Consul URL found.", "res": False}
            assert expected == result

    name = "name1"

    # no required arguments
    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            with patch.object(
                salt.modules.consul, "session_list", return_value=mock_result
            ):
                pytest.raises(
                    SaltInvocationError,
                    consul.agent_service_register,
                    consul_url=consul_url,
                )

    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            with patch.object(
                salt.modules.consul, "session_list", return_value=mock_result
            ):
                msg = "Service {} registered on agent."
                result = consul.agent_service_register(
                    consul_url=consul_url,
                    name=name,
                    script="test",
                    http="test",
                    ttl="test",
                    interval="test",
                )
                expected = {"message": msg.format(name), "res": True}
                assert expected == result

    with patch.object(salt.utils.http, "query", return_value=mock_http_result_false):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            with patch.object(
                salt.modules.consul, "session_list", return_value=mock_result
            ):
                msg = "Unable to register service {}."
                result = consul.agent_service_register(
                    consul_url=consul_url,
                    name=name,
                    script="test",
                    http="test",
                    ttl="test",
                    interval="test",
                )
                expected = {"message": msg.format(name), "res": False}
                assert expected == result


def test_agent_service_deregister():
    """
    Test consul agent service deregister
    """
    consul_url = "http://localhost:1313"
    key = "cluster/key"

    mock_result = "test"
    mock_http_result = {"status": 200, "dict": mock_result}
    mock_http_result_false = {"status": 204, "dict": mock_result}
    mock_url = MagicMock(return_value=consul_url)
    mock_nourl = MagicMock(return_value=None)

    # no consul url error
    with patch.dict(consul.__salt__, {"config.get": mock_nourl}):
        with patch.object(
            salt.modules.consul, "session_list", return_value=mock_result
        ):
            result = consul.agent_service_deregister(consul_url="")
            expected = {"message": "No Consul URL found.", "res": False}
            assert expected == result

    serviceid = "sid1"

    # no required arguments
    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            with patch.object(
                salt.modules.consul, "session_list", return_value=mock_result
            ):
                pytest.raises(
                    SaltInvocationError,
                    consul.agent_service_deregister,
                    consul_url=consul_url,
                )

    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            with patch.object(
                salt.modules.consul, "session_list", return_value=mock_result
            ):
                msg = "Service {} removed from agent."
                result = consul.agent_service_deregister(
                    consul_url=consul_url, serviceid=serviceid
                )
                expected = {"message": msg.format(serviceid), "res": True}
                assert expected == result

    with patch.object(salt.utils.http, "query", return_value=mock_http_result_false):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            with patch.object(
                salt.modules.consul, "session_list", return_value=mock_result
            ):
                msg = "Unable to remove service {}."
                result = consul.agent_service_deregister(
                    consul_url=consul_url, serviceid=serviceid
                )
                expected = {"message": msg.format(serviceid), "res": False}
                assert expected == result


def test_agent_service_maintenance():
    """
    Test consul agent service maintenance
    """
    consul_url = "http://localhost:1313"
    key = "cluster/key"

    mock_result = "test"
    mock_http_result = {"status": 200, "dict": mock_result}
    mock_http_result_false = {"status": 204, "dict": mock_result}
    mock_url = MagicMock(return_value=consul_url)
    mock_nourl = MagicMock(return_value=None)

    # no consul url error
    with patch.dict(consul.__salt__, {"config.get": mock_nourl}):
        with patch.object(
            salt.modules.consul, "session_list", return_value=mock_result
        ):
            result = consul.agent_service_maintenance(consul_url="")
            expected = {"message": "No Consul URL found.", "res": False}
            assert expected == result

    serviceid = "sid1"

    # no required arguments
    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            with patch.object(
                salt.modules.consul, "session_list", return_value=mock_result
            ):
                pytest.raises(
                    SaltInvocationError,
                    consul.agent_service_maintenance,
                    consul_url=consul_url,
                )

                # missing enable
                msg = 'Required parameter "enable" is missing.'
                result = consul.agent_service_maintenance(
                    consul_url=consul_url, serviceid=serviceid
                )
                expected = {"message": msg, "res": False}
                assert expected == result

    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            with patch.object(
                salt.modules.consul, "session_list", return_value=mock_result
            ):
                msg = "Service {} set in maintenance mode."
                result = consul.agent_service_maintenance(
                    consul_url=consul_url, serviceid=serviceid, enable=True
                )
                expected = {"message": msg.format(serviceid), "res": True}
                assert expected == result

    with patch.object(salt.utils.http, "query", return_value=mock_http_result_false):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            with patch.object(
                salt.modules.consul, "session_list", return_value=mock_result
            ):
                msg = "Unable to set service {} to maintenance mode."
                result = consul.agent_service_maintenance(
                    consul_url=consul_url, serviceid=serviceid, enable=True
                )
                expected = {"message": msg.format(serviceid), "res": False}
                assert expected == result


def test_session_create():
    """
    Test consul session create
    """
    consul_url = "http://localhost:1313"
    key = "cluster/key"

    mock_result = "test"
    mock_http_result = {"status": 200, "dict": mock_result}
    mock_http_result_false = {"status": 204, "dict": mock_result}
    mock_url = MagicMock(return_value=consul_url)
    mock_nourl = MagicMock(return_value=None)

    # no consul url error
    with patch.dict(consul.__salt__, {"config.get": mock_nourl}):
        with patch.object(
            salt.modules.consul, "session_list", return_value=mock_result
        ):
            result = consul.session_create(consul_url="")
            expected = {"message": "No Consul URL found.", "res": False}
            assert expected == result

    name = "name1"

    # no required arguments
    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            with patch.object(
                salt.modules.consul, "session_list", return_value=mock_result
            ):
                pytest.raises(
                    SaltInvocationError,
                    consul.session_create,
                    consul_url=consul_url,
                )

    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            with patch.object(
                salt.modules.consul, "session_list", return_value=mock_result
            ):
                msg = "Created session {}."
                result = consul.session_create(consul_url=consul_url, name=name)
                expected = {"message": msg.format(name), "res": True}
                assert expected == result

    with patch.object(salt.utils.http, "query", return_value=mock_http_result_false):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            with patch.object(
                salt.modules.consul, "session_list", return_value=mock_result
            ):
                msg = "Unable to create session {}."
                result = consul.session_create(consul_url=consul_url, name=name)
                expected = {"message": msg.format(name), "res": False}
                assert expected == result


def test_session_list():
    """
    Test consul session list
    """
    consul_url = "http://localhost:1313"
    key = "cluster/key"

    mock_result = "test"
    mock_http_result = {"status": 200, "dict": mock_result}
    mock_url = MagicMock(return_value=consul_url)
    mock_nourl = MagicMock(return_value=None)

    # no consul url error
    with patch.dict(consul.__salt__, {"config.get": mock_nourl}):
        result = consul.session_list(consul_url="")
        expected = {"message": "No Consul URL found.", "res": False}
        assert expected == result

    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            result = consul.session_list(consul_url=consul_url)
            expected = {"data": "test", "res": True}
            assert expected == result


def test_session_destroy():
    """
    Test consul session destroy
    """
    consul_url = "http://localhost:1313"
    key = "cluster/key"

    mock_result = "test"
    mock_http_result = {"status": 200, "dict": mock_result}
    mock_url = MagicMock(return_value=consul_url)
    mock_nourl = MagicMock(return_value=None)

    session = "sid1"
    name = "test"

    # no consul url error
    with patch.dict(consul.__salt__, {"config.get": mock_nourl}):
        result = consul.session_destroy(consul_url="")
        expected = {"message": "No Consul URL found.", "res": False}
        assert expected == result

    # no required arguments
    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            pytest.raises(
                SaltInvocationError,
                consul.session_destroy,
                consul_url=consul_url,
            )

    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            msg = "Destroyed Session {}."
            result = consul.session_destroy(
                consul_url=consul_url, session=session, name="test"
            )
            expected = {"message": msg.format(session), "res": True}
            assert expected == result


def test_session_info():
    """
    Test consul session info
    """
    consul_url = "http://localhost:1313"

    mock_result = "test"
    mock_http_result = {"status": 200, "dict": mock_result}
    mock_url = MagicMock(return_value=consul_url)
    mock_nourl = MagicMock(return_value=None)

    session = "sid1"

    # no consul url error
    with patch.dict(consul.__salt__, {"config.get": mock_nourl}):
        result = consul.session_info(consul_url="")
        expected = {"message": "No Consul URL found.", "res": False}
        assert expected == result

    # no required arguments
    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            pytest.raises(
                SaltInvocationError,
                consul.session_info,
                consul_url=consul_url,
            )

    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            result = consul.session_info(consul_url=consul_url, session=session)
            expected = {"data": "test", "res": True}
            assert expected == result


def test_catalog_register():
    """
    Test consul catalog register
    """
    consul_url = "http://localhost:1313"
    token = "randomtoken"
    node = "node1"
    address = "addres1"
    nodemeta = {
        "Cpu": "blah",
        "Cpu_num": "8",
        "Memory": "1024",
        "Os": "rhel8",
        "Osarch": "x86_64",
        "Kernel": "foo.bar",
        "Kernelrelease": "foo.release",
        "localhost": "localhost",
        "nodename": node,
        "os_family": "adams",
        "lsb_distrib_description": "distro",
        "master": "master",
    }
    nodemeta_kwargs = {
        "cpu": "blah",
        "num_cpus": "8",
        "mem": "1024",
        "oscode": "rhel8",
        "osarch": "x86_64",
        "kernel": "foo.bar",
        "kernelrelease": "foo.release",
        "localhost": "localhost",
        "nodename": node,
        "os_family": "adams",
        "lsb_distrib_description": "distro",
        "master": "master",
    }

    mock_result = "test"
    mock_http_result = {"status": 200, "dict": mock_result}
    mock_url = MagicMock(return_value=consul_url)
    mock_nourl = MagicMock(return_value=None)

    # no consul url error
    with patch.dict(consul.__salt__, {"config.get": mock_nourl}):
        result = consul.catalog_register(consul_url="")
        expected = {"message": "No Consul URL found.", "res": False}
        assert expected == result

    # no required arguments
    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            result = consul.catalog_register(consul_url=consul_url, token=token)
            expected = {
                "message": "Required argument node argument is missing.",
                "res": False,
            }
            assert expected == result

    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            result = consul.catalog_register(
                consul_url=consul_url,
                token=token,
                node=node,
                address=address,
                **nodemeta_kwargs,
            )
            expected = {
                "data": {"Address": address, "Node": node, "NodeMeta": nodemeta},
                "message": f"Catalog registration for {node} successful.",
                "res": True,
            }

            assert expected == result


def test_catalog_deregister():
    """
    Test consul catalog deregister
    """
    consul_url = "http://localhost:1313"
    token = "randomtoken"
    node = "node1"
    address = "addres1"
    serviceid = "server1"
    checkid = "check1"

    mock_result = "test"
    mock_http_result = {"status": 200, "dict": mock_result}
    mock_url = MagicMock(return_value=consul_url)
    mock_nourl = MagicMock(return_value=None)

    # no consul url error
    with patch.dict(consul.__salt__, {"config.get": mock_nourl}):
        result = consul.catalog_deregister(consul_url="")
        expected = {"message": "No Consul URL found.", "res": False}
        assert expected == result

    # no required arguments
    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            result = consul.catalog_deregister(consul_url=consul_url, token=token)
            expected = {"message": "Node argument required.", "res": False}
            assert expected == result

    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            result = consul.catalog_deregister(
                consul_url=consul_url,
                token=token,
                node=node,
                serviceid=serviceid,
                checkid=checkid,
            )
            expected = {
                "message": f"Catalog item {node} removed.",
                "res": True,
            }

            assert expected == result


def test_catalog_datacenters():
    """
    Test consul catalog datacenters
    """
    consul_url = "http://localhost:1313"
    token = "randomtoken"

    mock_result = "test"
    mock_http_result = {"status": 200, "dict": mock_result}
    mock_http_result_false = {"status": 204, "dict": mock_result}
    mock_url = MagicMock(return_value=consul_url)
    mock_nourl = MagicMock(return_value=None)

    # no consul url error
    with patch.dict(consul.__salt__, {"config.get": mock_nourl}):
        result = consul.catalog_datacenters(consul_url="")
        expected = {"message": "No Consul URL found.", "res": False}
        assert expected == result

    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            result = consul.catalog_datacenters(consul_url=consul_url, token=token)
            expected = {"data": "test", "res": True}
            assert expected == result


def test_catalog_nodes():
    """
    Test consul catalog nodes
    """
    consul_url = "http://localhost:1313"
    token = "randomtoken"

    mock_result = "test"
    mock_http_result = {"status": 200, "dict": mock_result}
    mock_http_result_false = {"status": 204, "dict": mock_result}
    mock_url = MagicMock(return_value=consul_url)
    mock_nourl = MagicMock(return_value=None)

    # no consul url error
    with patch.dict(consul.__salt__, {"config.get": mock_nourl}):
        result = consul.catalog_nodes(consul_url="")
        expected = {"message": "No Consul URL found.", "res": False}
        assert expected == result

    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            result = consul.catalog_nodes(consul_url=consul_url, token=token)
            expected = {"data": "test", "res": True}
            assert expected == result


def test_catalog_services():
    """
    Test consul catalog services
    """
    consul_url = "http://localhost:1313"
    token = "randomtoken"

    mock_result = "test"
    mock_http_result = {"status": 200, "dict": mock_result}
    mock_http_result_false = {"status": 204, "dict": mock_result}
    mock_url = MagicMock(return_value=consul_url)
    mock_nourl = MagicMock(return_value=None)

    # no consul url error
    with patch.dict(consul.__salt__, {"config.get": mock_nourl}):
        result = consul.catalog_services(consul_url="")
        expected = {"message": "No Consul URL found.", "res": False}
        assert expected == result

    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            result = consul.catalog_services(consul_url=consul_url, token=token)
            expected = {"data": "test", "res": True}
            assert expected == result


def test_catalog_service():
    """
    Test consul catalog service
    """
    consul_url = "http://localhost:1313"
    token = "randomtoken"
    service = "service"

    mock_result = "test"
    mock_http_result = {"status": 200, "dict": mock_result}
    mock_http_result_false = {"status": 204, "dict": mock_result}
    mock_url = MagicMock(return_value=consul_url)
    mock_nourl = MagicMock(return_value=None)

    # no consul url error
    with patch.dict(consul.__salt__, {"config.get": mock_nourl}):
        result = consul.catalog_service(consul_url="")
        expected = {"message": "No Consul URL found.", "res": False}
        assert expected == result

    # no required arguments
    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            pytest.raises(
                SaltInvocationError,
                consul.catalog_service,
                consul_url=consul_url,
            )

    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            result = consul.catalog_service(
                consul_url=consul_url, token=token, service=service
            )
            expected = {"data": "test", "res": True}
            assert expected == result


def test_catalog_node():
    """
    Test consul catalog node
    """
    consul_url = "http://localhost:1313"
    token = "randomtoken"
    node = "node"

    mock_result = "test"
    mock_http_result = {"status": 200, "dict": mock_result}
    mock_http_result_false = {"status": 204, "dict": mock_result}
    mock_url = MagicMock(return_value=consul_url)
    mock_nourl = MagicMock(return_value=None)

    # no consul url error
    with patch.dict(consul.__salt__, {"config.get": mock_nourl}):
        result = consul.catalog_node(consul_url="")
        expected = {"message": "No Consul URL found.", "res": False}
        assert expected == result

    # no required arguments
    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            pytest.raises(
                SaltInvocationError,
                consul.catalog_node,
                consul_url=consul_url,
            )

    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            result = consul.catalog_node(consul_url=consul_url, token=token, node=node)
            expected = {"data": "test", "res": True}
            assert expected == result


def test_health_node():
    """
    Test consul health node
    """
    consul_url = "http://localhost:1313"
    token = "randomtoken"
    node = "node"

    mock_result = "test"
    mock_http_result = {"status": 200, "dict": mock_result}
    mock_http_result_false = {"status": 204, "dict": mock_result}
    mock_url = MagicMock(return_value=consul_url)
    mock_nourl = MagicMock(return_value=None)

    # no consul url error
    with patch.dict(consul.__salt__, {"config.get": mock_nourl}):
        result = consul.health_node(consul_url="")
        expected = {"message": "No Consul URL found.", "res": False}
        assert expected == result

    # no required arguments
    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            pytest.raises(
                SaltInvocationError,
                consul.health_node,
                consul_url=consul_url,
            )

    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            result = consul.health_node(consul_url=consul_url, token=token, node=node)
            expected = {"data": "test", "res": True}
            assert expected == result


def test_health_checks():
    """
    Test consul health checks
    """
    consul_url = "http://localhost:1313"
    token = "randomtoken"
    service = "service"

    mock_result = "test"
    mock_http_result = {"status": 200, "dict": mock_result}
    mock_http_result_false = {"status": 204, "dict": mock_result}
    mock_url = MagicMock(return_value=consul_url)
    mock_nourl = MagicMock(return_value=None)

    # no consul url error
    with patch.dict(consul.__salt__, {"config.get": mock_nourl}):
        result = consul.health_checks(consul_url="")
        expected = {"message": "No Consul URL found.", "res": False}
        assert expected == result

    # no required arguments
    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            pytest.raises(
                SaltInvocationError,
                consul.health_checks,
                consul_url=consul_url,
            )

    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            result = consul.health_checks(
                consul_url=consul_url, token=token, service=service
            )
            expected = {"data": "test", "res": True}
            assert expected == result


def test_health_service():
    """
    Test consul health service
    """
    consul_url = "http://localhost:1313"
    token = "randomtoken"
    service = "service"

    mock_result = "test"
    mock_http_result = {"status": 200, "dict": mock_result}
    mock_http_result_false = {"status": 204, "dict": mock_result}
    mock_url = MagicMock(return_value=consul_url)
    mock_nourl = MagicMock(return_value=None)

    # no consul url error
    with patch.dict(consul.__salt__, {"config.get": mock_nourl}):
        result = consul.health_service(consul_url="")
        expected = {"message": "No Consul URL found.", "res": False}
        assert expected == result

    # no required arguments
    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            pytest.raises(
                SaltInvocationError,
                consul.health_service,
                consul_url=consul_url,
            )

    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            result = consul.health_service(
                consul_url=consul_url, token=token, service=service
            )
            expected = {"data": "test", "res": True}
            assert expected == result


def test_health_state():
    """
    Test consul health state
    """
    consul_url = "http://localhost:1313"
    token = "randomtoken"
    state = "state"

    mock_result = "test"
    mock_http_result = {"status": 200, "dict": mock_result}
    mock_http_result_false = {"status": 204, "dict": mock_result}
    mock_url = MagicMock(return_value=consul_url)
    mock_nourl = MagicMock(return_value=None)

    # no consul url error
    with patch.dict(consul.__salt__, {"config.get": mock_nourl}):
        result = consul.health_state(consul_url="")
        expected = {"message": "No Consul URL found.", "res": False}
        assert expected == result

    # no required arguments
    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            pytest.raises(
                SaltInvocationError,
                consul.health_state,
                consul_url=consul_url,
            )

    # state not in allowed states
    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            result = consul.health_state(
                consul_url=consul_url, token=token, state=state
            )
            expected = {
                "message": "State must be any, unknown, passing, warning, or critical.",
                "res": False,
            }
            assert expected == result

    state = "warning"
    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            result = consul.health_state(
                consul_url=consul_url, token=token, state=state
            )
            expected = {"data": "test", "res": True}
            assert expected == result


def test_status_leader():
    """
    Test consul status leader
    """
    consul_url = "http://localhost:1313"
    token = "randomtoken"

    mock_result = "test"
    mock_http_result = {"status": 200, "dict": mock_result}
    mock_http_result_false = {"status": 204, "dict": mock_result}
    mock_url = MagicMock(return_value=consul_url)
    mock_nourl = MagicMock(return_value=None)

    # no consul url error
    with patch.dict(consul.__salt__, {"config.get": mock_nourl}):
        result = consul.status_leader(consul_url="")
        expected = {"message": "No Consul URL found.", "res": False}
        assert expected == result

    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            result = consul.status_leader(consul_url=consul_url, token=token)
            expected = {"data": "test", "res": True}
            assert expected == result


def test_status_peers():
    """
    Test consul status peers
    """
    consul_url = "http://localhost:1313"
    token = "randomtoken"

    mock_result = "test"
    mock_http_result = {"status": 200, "dict": mock_result}
    mock_http_result_false = {"status": 204, "dict": mock_result}
    mock_url = MagicMock(return_value=consul_url)
    mock_nourl = MagicMock(return_value=None)

    # no consul url error
    with patch.dict(consul.__salt__, {"config.get": mock_nourl}):
        result = consul.status_peers(consul_url="")
        expected = {"message": "No Consul URL found.", "res": False}
        assert expected == result

    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            result = consul.status_peers(consul_url=consul_url, token=token)
            expected = {"data": "test", "res": True}
            assert expected == result


def test_acl_create():
    """
    Test consul acl create
    """
    consul_url = "http://localhost:1313"
    token = "randomtoken"
    name = "name1"

    mock_result = "test"
    mock_http_result = {"status": 200, "dict": mock_result}
    mock_http_result_false = {"status": 204, "dict": mock_result}
    mock_url = MagicMock(return_value=consul_url)
    mock_nourl = MagicMock(return_value=None)

    # no consul url error
    with patch.dict(consul.__salt__, {"config.get": mock_nourl}):
        result = consul.acl_create(consul_url="")
        expected = {"message": "No Consul URL found.", "res": False}
        assert expected == result

    # no required arguments
    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            pytest.raises(
                SaltInvocationError,
                consul.acl_create,
                consul_url=consul_url,
            )

    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            result = consul.acl_create(consul_url=consul_url, token=token, name=name)
            expected = {"message": f"ACL {name} created.", "res": True}
            assert expected == result


def test_acl_update():
    """
    Test consul acl update
    """
    consul_url = "http://localhost:1313"
    token = "randomtoken"
    name = "name1"
    aclid = "id1"

    mock_result = "test"
    mock_http_result = {"status": 200, "dict": mock_result}
    mock_http_result_false = {"status": 204, "dict": mock_result}
    mock_url = MagicMock(return_value=consul_url)
    mock_nourl = MagicMock(return_value=None)

    # no consul url error
    with patch.dict(consul.__salt__, {"config.get": mock_nourl}):
        result = consul.acl_update(consul_url="")
        expected = {"message": "No Consul URL found.", "res": False}
        assert expected == result

    # no required arguments
    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            result = consul.acl_update(consul_url=consul_url)
            expected = {
                "message": 'Required parameter "id" is missing.',
                "res": False,
            }
            assert expected == result

    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            pytest.raises(
                SaltInvocationError,
                consul.acl_update,
                consul_url=consul_url,
                id=aclid,
            )

    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            result = consul.acl_update(
                consul_url=consul_url, token=token, name=name, id=aclid
            )
            expected = {"message": f"ACL {name} created.", "res": True}
            assert expected == result


def test_acl_delete():
    """
    Test consul acl delete
    """
    consul_url = "http://localhost:1313"
    token = "randomtoken"
    name = "name1"
    aclid = "id1"

    mock_result = "test"
    mock_http_result = {"status": 200, "dict": mock_result}
    mock_http_result_false = {"status": 204, "dict": mock_result}
    mock_url = MagicMock(return_value=consul_url)
    mock_nourl = MagicMock(return_value=None)

    # no consul url error
    with patch.dict(consul.__salt__, {"config.get": mock_nourl}):
        result = consul.acl_delete(consul_url="")
        expected = {"message": "No Consul URL found.", "res": False}
        assert expected == result

    # no required arguments
    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            result = consul.acl_delete(consul_url=consul_url)
            expected = {
                "message": 'Required parameter "id" is missing.',
                "res": False,
            }
            assert expected == result

    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            result = consul.acl_delete(
                consul_url=consul_url, token=token, name=name, id=aclid
            )
            expected = {"message": f"ACL {aclid} deleted.", "res": True}
            assert expected == result


def test_acl_info():
    """
    Test consul acl info
    """
    consul_url = "http://localhost:1313"
    token = "randomtoken"
    name = "name1"
    aclid = "id1"

    mock_result = "test"
    mock_http_result = {"status": 200, "dict": mock_result}
    mock_http_result_false = {"status": 204, "dict": mock_result}
    mock_url = MagicMock(return_value=consul_url)
    mock_nourl = MagicMock(return_value=None)

    # no consul url error
    with patch.dict(consul.__salt__, {"config.get": mock_nourl}):
        result = consul.acl_info(consul_url="")
        expected = {"message": "No Consul URL found.", "res": False}
        assert expected == result

    # no required arguments
    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            result = consul.acl_info(consul_url=consul_url)
            expected = {
                "message": 'Required parameter "id" is missing.',
                "res": False,
            }
            assert expected == result

    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            result = consul.acl_info(
                consul_url=consul_url, token=token, name=name, id=aclid
            )
            expected = {"data": "test", "res": True}
            assert expected == result


def test_acl_clone():
    """
    Test consul acl clone
    """
    consul_url = "http://localhost:1313"
    token = "randomtoken"
    name = "name1"
    aclid = "id1"

    mock_result = aclid
    mock_http_result = {"status": 200, "dict": mock_result}
    mock_http_result_false = {"status": 204, "dict": mock_result}
    mock_url = MagicMock(return_value=consul_url)
    mock_nourl = MagicMock(return_value=None)

    # no consul url error
    with patch.dict(consul.__salt__, {"config.get": mock_nourl}):
        result = consul.acl_clone(consul_url="")
        expected = {"message": "No Consul URL found.", "res": False}
        assert expected == result

    # no required arguments
    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            result = consul.acl_clone(consul_url=consul_url)
            expected = {
                "message": 'Required parameter "id" is missing.',
                "res": False,
            }
            assert expected == result

    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            result = consul.acl_clone(
                consul_url=consul_url, token=token, name=name, id=aclid
            )
            expected = {
                "ID": aclid,
                "message": f"ACL {name} cloned.",
                "res": True,
            }
            assert expected == result


def test_acl_list():
    """
    Test consul acl list
    """
    consul_url = "http://localhost:1313"
    token = "randomtoken"
    name = "name1"
    aclid = "id1"

    mock_result = aclid
    mock_http_result = {"status": 200, "dict": mock_result}
    mock_http_result_false = {"status": 204, "dict": mock_result}
    mock_url = MagicMock(return_value=consul_url)
    mock_nourl = MagicMock(return_value=None)

    # no consul url error
    with patch.dict(consul.__salt__, {"config.get": mock_nourl}):
        result = consul.acl_list(consul_url="")
        expected = {"message": "No Consul URL found.", "res": False}
        assert expected == result

    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            result = consul.acl_list(
                consul_url=consul_url, token=token, name=name, id=aclid
            )
            expected = {"data": "id1", "res": True}
            assert expected == result


def test_event_fire():
    """
    Test consul event fire
    """
    consul_url = "http://localhost:1313"
    token = "randomtoken"
    name = "name1"

    mock_result = "test"
    mock_http_result = {"status": 200, "dict": mock_result}
    mock_http_result_false = {"status": 204, "dict": mock_result}
    mock_url = MagicMock(return_value=consul_url)
    mock_nourl = MagicMock(return_value=None)

    # no consul url error
    with patch.dict(consul.__salt__, {"config.get": mock_nourl}):
        result = consul.event_fire(consul_url="")
        expected = {"message": "No Consul URL found.", "res": False}
        assert expected == result

    # no required arguments
    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            pytest.raises(
                SaltInvocationError,
                consul.event_fire,
                consul_url=consul_url,
            )

    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            result = consul.event_fire(consul_url=consul_url, token=token, name=name)
            expected = {
                "data": "test",
                "message": f"Event {name} fired.",
                "res": True,
            }
            assert expected == result


def test_event_list():
    """
    Test consul event list
    """
    consul_url = "http://localhost:1313"
    token = "randomtoken"
    name = "name1"

    mock_result = "test"
    mock_http_result = {"status": 200, "dict": mock_result}
    mock_http_result_false = {"status": 204, "dict": mock_result}
    mock_url = MagicMock(return_value=consul_url)
    mock_nourl = MagicMock(return_value=None)

    # no consul url error
    with patch.dict(consul.__salt__, {"config.get": mock_nourl}):
        result = consul.event_list(consul_url="")
        expected = {"message": "No Consul URL found.", "res": False}
        assert expected == result

    # no required arguments
    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            pytest.raises(
                SaltInvocationError,
                consul.event_list,
                consul_url=consul_url,
            )

    with patch.object(salt.utils.http, "query", return_value=mock_http_result):
        with patch.dict(consul.__salt__, {"config.get": mock_url}):
            result = consul.event_list(consul_url=consul_url, token=token, name=name)
            expected = {"data": "test", "res": True}
            assert expected == result



"""
Zookeeper Module
~~~~~~~~~~~~~~~~
:maintainer:    SaltStack
:maturity:      new
:platform:      all
:depends:       kazoo

.. versionadded:: 2018.3.0

Configuration
=============

:configuration: This module is not usable until the following are specified
    either in a pillar or in the minion's config file:

    .. code-block:: yaml

        zookeeper:
          hosts: zoo1,zoo2,zoo3
          default_acl:
            - username: daniel
              password: test
              read: true
              write: true
              create: true
              delete: true
              admin: true
          username: daniel
          password: test

    If configuration for multiple zookeeper environments is required, they can
    be set up as different configuration profiles. For example:

    .. code-block:: yaml

        zookeeper:
          prod:
            hosts: zoo1,zoo2,zoo3
            default_acl:
              - username: daniel
                password: test
                read: true
                write: true
                create: true
                delete: true
                admin: true
            username: daniel
            password: test
          dev:
            hosts:
              - dev1
              - dev2
              - dev3
            default_acl:
              - username: daniel
                password: test
                read: true
                write: true
                create: true
                delete: true
                admin: true
            username: daniel
            password: test
"""

# Import Salt libraries
import salt.utils.stringutils

# Import python libraries
try:
    import kazoo.client
    import kazoo.security

    HAS_KAZOO = True
except ImportError:
    HAS_KAZOO = False


__virtualname__ = "zookeeper"


def __virtual__():
    if HAS_KAZOO:
        return __virtualname__
    return (False, "Missing dependency: kazoo")


def _get_zk_conn(profile=None, **connection_args):
    if profile:
        prefix = "zookeeper:" + profile
    else:
        prefix = "zookeeper"

    def get(key, default=None):
        """
        look in connection_args first, then default to config file
        """
        return connection_args.get(key) or __salt__["config.get"](
            ":".join([prefix, key]), default
        )

    hosts = get("hosts", "127.0.0.1:2181")
    scheme = get("scheme", None)
    username = get("username", None)
    password = get("password", None)
    default_acl = get("default_acl", None)

    if isinstance(hosts, list):
        hosts = ",".join(hosts)

    if username is not None and password is not None and scheme is None:
        scheme = "digest"

    auth_data = None
    if scheme and username and password:
        auth_data = [(scheme, ":".join([username, password]))]

    if default_acl is not None:
        if isinstance(default_acl, list):
            default_acl = [make_digest_acl(**acl) for acl in default_acl]
        else:
            default_acl = [make_digest_acl(**default_acl)]

    __context__.setdefault("zkconnection", {}).setdefault(
        profile or hosts,
        kazoo.client.KazooClient(
            hosts=hosts, default_acl=default_acl, auth_data=auth_data
        ),
    )

    if not __context__["zkconnection"][profile or hosts].connected:
        __context__["zkconnection"][profile or hosts].start()

    return __context__["zkconnection"][profile or hosts]


def create(
    path,
    value="",
    acls=None,
    ephemeral=False,
    sequence=False,
    makepath=False,
    profile=None,
    hosts=None,
    scheme=None,
    username=None,
    password=None,
    default_acl=None,
):
    """
    Create Znode

    path
        path of znode to create

    value
        value to assign to znode (Default: '')

    acls
        list of acl dictionaries to be assigned (Default: None)

    ephemeral
        indicate node is ephemeral (Default: False)

    sequence
        indicate node is suffixed with a unique index (Default: False)

    makepath
        Create parent paths if they do not exist (Default: False)

    profile
        Configured Zookeeper profile to authenticate with (Default: None)

    hosts
        Lists of Zookeeper Hosts (Default: '127.0.0.1:2181)

    scheme
        Scheme to authenticate with (Default: 'digest')

    username
        Username to authenticate (Default: None)

    password
        Password to authenticate (Default: None)

    default_acl
        Default acls to assign if a node is created in this connection (Default: None)

    CLI Example:

    .. code-block:: bash

        salt minion1 zookeeper.create /test/name daniel profile=prod

    """
    if acls is None:
        acls = []
    acls = [make_digest_acl(**acl) for acl in acls]
    conn = _get_zk_conn(
        profile=profile,
        hosts=hosts,
        scheme=scheme,
        username=username,
        password=password,
        default_acl=default_acl,
    )
    return conn.create(
        path,
        salt.utils.stringutils.to_bytes(value),
        acls,
        ephemeral,
        sequence,
        makepath,
    )


def ensure_path(
    path,
    acls=None,
    profile=None,
    hosts=None,
    scheme=None,
    username=None,
    password=None,
    default_acl=None,
):
    """
    Ensure Znode path exists

    path
        Parent path to create

    acls
        list of acls dictionaries to be assigned (Default: None)

    profile
        Configured Zookeeper profile to authenticate with (Default: None)

    hosts
        Lists of Zookeeper Hosts (Default: '127.0.0.1:2181)

    scheme
        Scheme to authenticate with (Default: 'digest')

    username
        Username to authenticate (Default: None)

    password
        Password to authenticate (Default: None)

    default_acl
        Default acls to assign if a node is created in this connection (Default: None)

    CLI Example:

    .. code-block:: bash

        salt minion1 zookeeper.ensure_path /test/name profile=prod

    """
    if acls is None:
        acls = []
    acls = [make_digest_acl(**acl) for acl in acls]
    conn = _get_zk_conn(
        profile=profile,
        hosts=hosts,
        scheme=scheme,
        username=username,
        password=password,
        default_acl=default_acl,
    )
    return conn.ensure_path(path, acls)


def exists(
    path,
    profile=None,
    hosts=None,
    scheme=None,
    username=None,
    password=None,
    default_acl=None,
):
    """
    Check if path exists

    path
        path to check

    profile
        Configured Zookeeper profile to authenticate with (Default: None)

    hosts
        Lists of Zookeeper Hosts (Default: '127.0.0.1:2181)

    scheme
        Scheme to authenticate with (Default: 'digest')

    username
        Username to authenticate (Default: None)

    password
        Password to authenticate (Default: None)

    default_acl
        Default acls to assign if a node is created in this connection (Default: None)

    CLI Example:

    .. code-block:: bash

        salt minion1 zookeeper.exists /test/name profile=prod

    """
    conn = _get_zk_conn(
        profile=profile,
        hosts=hosts,
        scheme=scheme,
        username=username,
        password=password,
        default_acl=default_acl,
    )
    return bool(conn.exists(path))


def get(
    path,
    profile=None,
    hosts=None,
    scheme=None,
    username=None,
    password=None,
    default_acl=None,
):
    """
    Get value saved in znode

    path
        path to check

    profile
        Configured Zookeeper profile to authenticate with (Default: None)

    hosts
        Lists of Zookeeper Hosts (Default: '127.0.0.1:2181)

    scheme
        Scheme to authenticate with (Default: 'digest')

    username
        Username to authenticate (Default: None)

    password
        Password to authenticate (Default: None)

    default_acl
        Default acls to assign if a node is created in this connection (Default: None)

    CLI Example:

    .. code-block:: bash

        salt minion1 zookeeper.get /test/name profile=prod

    """
    conn = _get_zk_conn(
        profile=profile,
        hosts=hosts,
        scheme=scheme,
        username=username,
        password=password,
        default_acl=default_acl,
    )
    ret, _ = conn.get(path)
    return salt.utils.stringutils.to_str(ret)


def get_children(
    path,
    profile=None,
    hosts=None,
    scheme=None,
    username=None,
    password=None,
    default_acl=None,
):
    """
    Get children in znode path

    path
        path to check

    profile
        Configured Zookeeper profile to authenticate with (Default: None)

    hosts
        Lists of Zookeeper Hosts (Default: '127.0.0.1:2181)

    scheme
        Scheme to authenticate with (Default: 'digest')

    username
        Username to authenticate (Default: None)

    password
        Password to authenticate (Default: None)

    default_acl
        Default acls to assign if a node is created in this connection (Default: None)

    CLI Example:

    .. code-block:: bash

        salt minion1 zookeeper.get_children /test profile=prod

    """
    conn = _get_zk_conn(
        profile=profile,
        hosts=hosts,
        scheme=scheme,
        username=username,
        password=password,
        default_acl=default_acl,
    )
    ret = conn.get_children(path)
    return ret or []


def set(
    path,
    value,
    version=-1,
    profile=None,
    hosts=None,
    scheme=None,
    username=None,
    password=None,
    default_acl=None,
):
    """
    Update znode with new value

    path
        znode to update

    value
        value to set in znode

    version
        only update znode if version matches (Default: -1 (always matches))

    profile
        Configured Zookeeper profile to authenticate with (Default: None)

    hosts
        Lists of Zookeeper Hosts (Default: '127.0.0.1:2181)

    scheme
        Scheme to authenticate with (Default: 'digest')

    username
        Username to authenticate (Default: None)

    password
        Password to authenticate (Default: None)

    default_acl
        Default acls to assign if a node is created in this connection (Default: None)

    CLI Example:

    .. code-block:: bash

        salt minion1 zookeeper.set /test/name gtmanfred profile=prod

    """
    conn = _get_zk_conn(
        profile=profile,
        hosts=hosts,
        scheme=scheme,
        username=username,
        password=password,
        default_acl=default_acl,
    )
    return conn.set(path, salt.utils.stringutils.to_bytes(value), version=version)


def get_acls(
    path,
    profile=None,
    hosts=None,
    scheme=None,
    username=None,
    password=None,
    default_acl=None,
):
    """
    Get acls on a znode

    path
        path to znode

    profile
        Configured Zookeeper profile to authenticate with (Default: None)

    hosts
        Lists of Zookeeper Hosts (Default: '127.0.0.1:2181)

    scheme
        Scheme to authenticate with (Default: 'digest')

    username
        Username to authenticate (Default: None)

    password
        Password to authenticate (Default: None)

    default_acl
        Default acls to assign if a node is created in this connection (Default: None)

    CLI Example:

    .. code-block:: bash

        salt minion1 zookeeper.get_acls /test/name profile=prod

    """
    conn = _get_zk_conn(
        profile=profile,
        hosts=hosts,
        scheme=scheme,
        username=username,
        password=password,
        default_acl=default_acl,
    )
    return conn.get_acls(path)[0]


def set_acls(
    path,
    acls,
    version=-1,
    profile=None,
    hosts=None,
    scheme=None,
    username=None,
    password=None,
    default_acl=None,
):
    """
    Set acls on a znode

    path
        path to znode

    acls
        list of acl dictionaries to set on the znode

    version
        only set acls if version matches (Default: -1 (always matches))

    profile
        Configured Zookeeper profile to authenticate with (Default: None)

    hosts
        Lists of Zookeeper Hosts (Default: '127.0.0.1:2181)

    scheme
        Scheme to authenticate with (Default: 'digest')

    username
        Username to authenticate (Default: None)

    password
        Password to authenticate (Default: None)

    default_acl
        Default acls to assign if a node is created in this connection (Default: None)

    CLI Example:

    .. code-block:: bash

        salt minion1 zookeeper.set_acls /test/name acls='[{"username": "gtmanfred", "password": "test", "all": True}]' profile=prod

    """
    conn = _get_zk_conn(
        profile=profile,
        hosts=hosts,
        scheme=scheme,
        username=username,
        password=password,
        default_acl=default_acl,
    )
    if acls is None:
        acls = []
    acls = [make_digest_acl(**acl) for acl in acls]
    conn = _get_zk_conn(
        profile=profile,
        hosts=hosts,
        scheme=scheme,
        username=username,
        password=password,
        default_acl=default_acl,
    )
    return conn.set_acls(path, acls, version)


def delete(
    path,
    version=-1,
    recursive=False,
    profile=None,
    hosts=None,
    scheme=None,
    username=None,
    password=None,
    default_acl=None,
):
    """
    Delete znode

    path
        path to znode

    version
        only delete if version matches (Default: -1 (always matches))

    profile
        Configured Zookeeper profile to authenticate with (Default: None)

    hosts
        Lists of Zookeeper Hosts (Default: '127.0.0.1:2181)

    scheme
        Scheme to authenticate with (Default: 'digest')

    username
        Username to authenticate (Default: None)

    password
        Password to authenticate (Default: None)

    default_acl
        Default acls to assign if a node is created in this connection (Default: None)

    CLI Example:

    .. code-block:: bash

        salt minion1 zookeeper.delete /test/name profile=prod

    """
    conn = _get_zk_conn(
        profile=profile,
        hosts=hosts,
        scheme=scheme,
        username=username,
        password=password,
        default_acl=default_acl,
    )
    return conn.delete(path, version, recursive)


def make_digest_acl(
    username,
    password,
    read=False,
    write=False,
    create=False,
    delete=False,
    admin=False,
    allperms=False,
):
    """
    Generate acl object

    .. note:: This is heavily used in the zookeeper state and probably is not useful as a cli module

    username
        username of acl

    password
        plain text password of acl

    read
        read acl

    write
        write acl

    create
        create acl

    delete
        delete acl

    admin
        admin acl

    allperms
        set all other acls to True

    CLI Example:

    .. code-block:: bash

        salt minion1 zookeeper.make_digest_acl username=daniel password=mypass allperms=True
    """
    return kazoo.security.make_digest_acl(
        username, password, read, write, create, delete, admin, allperms
    )



"""
Test cases for salt.states.openvswitch_db.
"""

import pytest

import salt.states.openvswitch_db as openvswitch_db
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {openvswitch_db: {"__opts__": {"test": False}}}


def test_managed_different_entry_present():
    """
    Test managed function.

    This tests the case where there already is an entry, but it does not match.
    """
    get_mock = MagicMock(return_value="01:02:03:04:05:06")
    set_mock = MagicMock(return_value=None)
    with patch.dict(
        openvswitch_db.__salt__,
        {"openvswitch.db_get": get_mock, "openvswitch.db_set": set_mock},
    ):
        ret = openvswitch_db.managed(
            name="br0", table="Interface", data={"mac": "01:02:03:04:05:07"}
        )
        get_mock.assert_called_with("Interface", "br0", "mac", True)
        set_mock.assert_called_with("Interface", "br0", "mac", "01:02:03:04:05:07")
        assert ret["result"] is True
        assert ret["changes"] == {
            "mac": {"old": "01:02:03:04:05:06", "new": "01:02:03:04:05:07"}
        }


def test_managed_matching_entry_present():
    """
    Test managed function.

    This tests the case where there already is a matching entry.
    """
    get_mock = MagicMock(return_value="01:02:03:04:05:06")
    set_mock = MagicMock(return_value=None)
    with patch.dict(
        openvswitch_db.__salt__,
        {"openvswitch.db_get": get_mock, "openvswitch.db_set": set_mock},
    ):
        ret = openvswitch_db.managed(
            name="br0", table="Interface", data={"mac": "01:02:03:04:05:06"}
        )
        get_mock.assert_called_with("Interface", "br0", "mac", True)
        set_mock.assert_not_called()
        assert ret["result"] is True
        assert "changes" not in ret or not ret["changes"]


def test_managed_no_entry_present():
    """
    Test managed function.

    This tests the case where there is no entry yet.
    """
    get_mock = MagicMock(return_value="01:02:03:04:05:06")
    set_mock = MagicMock(return_value=None)
    with patch.dict(
        openvswitch_db.__salt__,
        {"openvswitch.db_get": get_mock, "openvswitch.db_set": set_mock},
    ):
        ret = openvswitch_db.managed(
            name="br0", table="Interface", data={"mac": "01:02:03:04:05:07"}
        )
        get_mock.assert_called_with("Interface", "br0", "mac", True)
        set_mock.assert_called_with("Interface", "br0", "mac", "01:02:03:04:05:07")
        assert ret["result"] is True
        assert ret["changes"] == {
            "mac": {"old": "01:02:03:04:05:06", "new": "01:02:03:04:05:07"}
        }



"""
This module allows SPM to use sqlite3 as the backend for SPM's package database.

.. versionadded:: 2015.8.0
"""

import datetime
import logging
import os
import sqlite3
from sqlite3 import OperationalError

# Get logging started
log = logging.getLogger(__name__)


def init():
    """
    Get an sqlite3 connection, and initialize the package database if necessary
    """
    if not os.path.exists(__opts__["spm_cache_dir"]):
        log.debug("Creating SPM cache directory at %s", __opts__["spm_db"])
        os.makedirs(__opts__["spm_cache_dir"])

    if not os.path.exists(__opts__["spm_db"]):
        log.debug("Creating new package database at %s", __opts__["spm_db"])

    sqlite3.enable_callback_tracebacks(True)
    conn = sqlite3.connect(__opts__["spm_db"], isolation_level=None)

    try:
        conn.execute("SELECT count(*) FROM packages")
    except OperationalError:
        conn.execute(
            """CREATE TABLE packages (
            package text,
            version text,
            release text,
            installed text,
            os text,
            os_family text,
            dependencies text,
            os_dependencies text,
            os_family_dependencies text,
            summary text,
            description text
        )"""
        )

    try:
        conn.execute("SELECT count(*) FROM files")
    except OperationalError:
        conn.execute(
            """CREATE TABLE files (
            package text,
            path text,
            size real,
            mode text,
            sum text,
            major text,
            minor text,
            linkname text,
            linkpath text,
            uname text,
            gname text,
            mtime text
        )"""
        )

    return conn


def info(package, conn=None):
    """
    List info for a package
    """
    close = False
    if conn is None:
        close = True
        conn = init()

    fields = (
        "package",
        "version",
        "release",
        "installed",
        "os",
        "os_family",
        "dependencies",
        "os_dependencies",
        "os_family_dependencies",
        "summary",
        "description",
    )
    data = conn.execute(
        "SELECT {} FROM packages WHERE package=?".format(",".join(fields)), (package,)
    )
    row = data.fetchone()
    if close:
        conn.close()
    if not row:
        return None

    formula_def = dict(list(zip(fields, row)))
    formula_def["name"] = formula_def["package"]

    return formula_def


def list_packages(conn=None):
    """
    List files for an installed package
    """
    close = False
    if conn is None:
        close = True
        conn = init()

    ret = []
    data = conn.execute("SELECT package FROM packages")
    for pkg in data.fetchall():
        ret.append(pkg)

    if close:
        conn.close()
    return ret


def list_files(package, conn=None):
    """
    List files for an installed package
    """
    close = False
    if conn is None:
        close = True
        conn = init()

    data = conn.execute("SELECT package FROM packages WHERE package=?", (package,))
    if not data.fetchone():
        if close:
            conn.close()
        return None

    ret = []
    data = conn.execute("SELECT path, sum FROM files WHERE package=?", (package,))
    for file_ in data.fetchall():
        ret.append(file_)
    if close:
        conn.close()

    return ret


def register_pkg(name, formula_def, conn=None):
    """
    Register a package in the package database
    """
    close = False
    if conn is None:
        close = True
        conn = init()

    conn.execute(
        "INSERT INTO packages VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)",
        (
            name,
            formula_def["version"],
            formula_def["release"],
            datetime.datetime.utcnow().strftime("%a, %d %b %Y %H:%M:%S GMT"),
            formula_def.get("os", None),
            formula_def.get("os_family", None),
            formula_def.get("dependencies", None),
            formula_def.get("os_dependencies", None),
            formula_def.get("os_family_dependencies", None),
            formula_def["summary"],
            formula_def["description"],
        ),
    )
    if close:
        conn.close()


def register_file(name, member, path, digest="", conn=None):
    """
    Register a file in the package database
    """
    close = False
    if conn is None:
        close = True
        conn = init()

    conn.execute(
        "INSERT INTO files VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)",
        (
            name,
            f"{path}/{member.path}",
            member.size,
            member.mode,
            digest,
            member.devmajor,
            member.devminor,
            member.linkname,
            member.linkpath,
            member.uname,
            member.gname,
            member.mtime,
        ),
    )
    if close:
        conn.close()


def unregister_pkg(name, conn=None):
    """
    Unregister a package from the package database
    """
    if conn is None:
        conn = init()

    conn.execute("DELETE FROM packages WHERE package=?", (name,))


def unregister_file(path, pkg=None, conn=None):  # pylint: disable=W0612
    """
    Unregister a file from the package database
    """
    close = False
    if conn is None:
        close = True
        conn = init()

    conn.execute("DELETE FROM files WHERE path=?", (path,))
    if close:
        conn.close()


def db_exists(db_):
    """
    Check to see whether the file already exists
    """
    return os.path.exists(db_)



"""
    :codeauthor: Pedro Algarvio (pedro@algarvio.me)


    tests.unit.utils.filebuffer_test
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

import os

from salt.utils.filebuffer import BufferedReader, InvalidFileMode
from tests.support.paths import BASE_FILES
from tests.support.unit import TestCase


class TestFileBuffer(TestCase):
    def test_read_only_mode(self):
        with self.assertRaises(InvalidFileMode):
            BufferedReader("/tmp/foo", mode="a")

        with self.assertRaises(InvalidFileMode):
            BufferedReader("/tmp/foo", mode="ab")

        with self.assertRaises(InvalidFileMode):
            BufferedReader("/tmp/foo", mode="w")

        with self.assertRaises(InvalidFileMode):
            BufferedReader("/tmp/foo", mode="wb")

    def test_issue_51309(self):
        """
        https://github.com/saltstack/salt/issues/51309
        """
        file_name = os.path.join(BASE_FILES, "grail", "scene33")

        def find_value(text):
            stripped_text = text.strip()
            try:
                with BufferedReader(file_name) as breader:
                    for chunk in breader:
                        if stripped_text in chunk:
                            return True
                return False
            except OSError:
                return False

        self.assertTrue(find_value("We have the Holy Hand Grenade"))



import pytest
from pytestskipmarkers.utils import platform

import salt.modules.postgres as postgres
import salt.states.postgres_group as postgres_group
from tests.support.mock import create_autospec, patch


@pytest.fixture(name="db_args")
def fixture_db_args():
    return {
        "runas": None,
        "host": None,
        "port": None,
        "maintenance_db": None,
        "user": None,
        "password": None,
    }


@pytest.fixture(name="md5_pw")
def fixture_md5_pw():
    if platform.is_fips_enabled():
        pytest.skip("Test cannot run on a FIPS enabled platform")
    # 'md5' + md5('password' + 'groupname')
    return "md58b14c378fab8ef0dc227f4e6d6787a87"


@pytest.fixture(name="existing_group")
def fixture_existing_group():
    return {
        "superuser": False,
        "inherits privileges": True,
        "can create roles": False,
        "can create databases": False,
        "can update system catalogs": None,
        "can login": False,
        "replication": False,
        "connections": None,
        "expiry time": None,
        "defaults variables": "",
        "password": "",
        "groups": [],
    }


@pytest.fixture(name="test_mode")
def fixture_test_mode():
    with patch.dict(postgres_group.__opts__, {"test": True}):
        yield


@pytest.fixture
def mocks():
    return {
        "postgres.role_get": create_autospec(postgres.role_get, return_value=None),
        "postgres.user_exists": create_autospec(
            postgres.user_exists, return_value=False
        ),
        "postgres.group_create": create_autospec(
            postgres.group_create, return_value=True
        ),
        "postgres.group_update": create_autospec(
            postgres.group_update, return_value=True
        ),
        "postgres.group_remove": create_autospec(
            postgres.group_remove, return_value=True
        ),
    }


@pytest.fixture
def configure_loader_modules(mocks):
    return {
        postgres_group: {"__opts__": {"test": False}, "__salt__": mocks},
        postgres: {"__opts__": {"test": False}},
    }


# ==========
# postgres_group.present
# ==========


@pytest.mark.skip_on_fips_enabled_platform
def test_present_create_basic(mocks, db_args):
    assert postgres_group.present("groupname") == {
        "name": "groupname",
        "result": True,
        "changes": {"groupname": "Present"},
        "comment": "The group groupname has been created",
    }
    mocks["postgres.role_get"].assert_called_once_with(
        "groupname", return_password=True, **db_args
    )
    mocks["postgres.group_create"].assert_called_once_with(
        groupname="groupname",
        createdb=None,
        createroles=None,
        encrypted="md5",
        superuser=None,
        login=None,
        inherit=None,
        replication=None,
        rolepassword=None,
        groups=None,
        **db_args
    )
    mocks["postgres.group_update"].assert_not_called()


@pytest.mark.usefixtures("test_mode")
def test_present_create_basic_test(mocks, db_args):
    assert postgres_group.present("groupname") == {
        "name": "groupname",
        "result": None,
        "changes": {},
        "comment": "Group groupname is set to be created",
    }
    mocks["postgres.role_get"].assert_called_once_with(
        "groupname", return_password=True, **db_args
    )
    mocks["postgres.group_create"].assert_not_called()
    mocks["postgres.group_update"].assert_not_called()


def test_present_exists_basic(mocks, existing_group, db_args):
    mocks["postgres.role_get"].return_value = existing_group

    assert postgres_group.present("groupname") == {
        "name": "groupname",
        "result": True,
        "changes": {},
        "comment": "Group groupname is already present",
    }
    mocks["postgres.role_get"].assert_called_once_with(
        "groupname", return_password=True, **db_args
    )
    mocks["postgres.group_create"].assert_not_called()
    mocks["postgres.group_update"].assert_not_called()


def test_present_create_basic_error(mocks, db_args):
    mocks["postgres.group_create"].return_value = False

    assert postgres_group.present("groupname") == {
        "name": "groupname",
        "result": False,
        "changes": {},
        "comment": "Failed to create group groupname",
    }
    mocks["postgres.role_get"].assert_called_once_with(
        "groupname", return_password=True, **db_args
    )
    mocks["postgres.group_create"].assert_called_once()
    mocks["postgres.group_update"].assert_not_called()


def test_present_change_option(mocks, existing_group, db_args):
    mocks["postgres.role_get"].return_value = existing_group

    assert postgres_group.present("groupname", replication=True) == {
        "name": "groupname",
        "result": True,
        "changes": {"groupname": {"replication": True}},
        "comment": "The group groupname has been updated",
    }

    mocks["postgres.role_get"].assert_called_once()
    mocks["postgres.group_create"].assert_not_called()
    mocks["postgres.group_update"].assert_called_once_with(
        groupname="groupname",
        createdb=None,
        createroles=None,
        encrypted="md5",
        superuser=None,
        login=None,
        inherit=None,
        replication=True,
        rolepassword=None,
        groups=None,
        **db_args
    )


def test_present_create_md5_password(mocks, md5_pw, db_args):
    assert postgres_group.present("groupname", password="password", encrypted=True) == {
        "name": "groupname",
        "result": True,
        "changes": {"groupname": "Present"},
        "comment": "The group groupname has been created",
    }
    mocks["postgres.role_get"].assert_called_once()
    mocks["postgres.group_create"].assert_called_once_with(
        groupname="groupname",
        createdb=None,
        createroles=None,
        encrypted=True,
        superuser=None,
        login=None,
        inherit=None,
        replication=None,
        rolepassword=md5_pw,
        groups=None,
        **db_args
    )
    mocks["postgres.group_update"].assert_not_called()


def test_present_create_plain_password(mocks, db_args):
    assert postgres_group.present(
        "groupname", password="password", encrypted=False
    ) == {
        "name": "groupname",
        "result": True,
        "changes": {"groupname": "Present"},
        "comment": "The group groupname has been created",
    }
    mocks["postgres.role_get"].assert_called_once()
    mocks["postgres.group_create"].assert_called_once_with(
        groupname="groupname",
        createdb=None,
        createroles=None,
        encrypted=False,
        superuser=None,
        login=None,
        inherit=None,
        replication=None,
        rolepassword="password",
        groups=None,
        **db_args
    )
    mocks["postgres.group_update"].assert_not_called()


def test_present_create_md5_password_default_plain(mocks, monkeypatch, md5_pw, db_args):
    monkeypatch.setattr(postgres, "_DEFAULT_PASSWORDS_ENCRYPTION", False)
    test_present_create_md5_password(mocks, md5_pw, db_args)


def test_present_create_md5_password_default_encrypted(
    mocks, monkeypatch, md5_pw, db_args
):
    monkeypatch.setattr(postgres, "_DEFAULT_PASSWORDS_ENCRYPTION", True)

    assert postgres_group.present("groupname", password="password") == {
        "name": "groupname",
        "result": True,
        "changes": {"groupname": "Present"},
        "comment": "The group groupname has been created",
    }
    mocks["postgres.role_get"].assert_called_once()
    mocks["postgres.group_create"].assert_called_once_with(
        groupname="groupname",
        createdb=None,
        createroles=None,
        encrypted=True,
        superuser=None,
        login=None,
        inherit=None,
        replication=None,
        rolepassword=md5_pw,
        groups=None,
        **db_args
    )
    mocks["postgres.group_update"].assert_not_called()


def test_present_create_md5_prehashed(mocks, md5_pw, db_args):
    assert postgres_group.present("groupname", password=md5_pw, encrypted=True) == {
        "name": "groupname",
        "result": True,
        "changes": {"groupname": "Present"},
        "comment": "The group groupname has been created",
    }
    mocks["postgres.role_get"].assert_called_once()
    mocks["postgres.group_create"].assert_called_once_with(
        groupname="groupname",
        createdb=None,
        createroles=None,
        encrypted=True,
        superuser=None,
        login=None,
        inherit=None,
        replication=None,
        rolepassword=md5_pw,
        groups=None,
        **db_args
    )
    mocks["postgres.group_update"].assert_not_called()


def test_present_md5_matches(mocks, existing_group, md5_pw):
    existing_group["password"] = md5_pw
    mocks["postgres.role_get"].return_value = existing_group

    assert postgres_group.present("groupname", password="password", encrypted=True) == {
        "name": "groupname",
        "result": True,
        "changes": {},
        "comment": "Group groupname is already present",
    }
    mocks["postgres.role_get"].assert_called_once()
    mocks["postgres.group_create"].assert_not_called()
    mocks["postgres.group_update"].assert_not_called()


def test_present_md5_matches_prehashed(mocks, existing_group, md5_pw):
    existing_group["password"] = md5_pw
    mocks["postgres.role_get"].return_value = existing_group

    assert postgres_group.present("groupname", password=md5_pw, encrypted=True) == {
        "name": "groupname",
        "result": True,
        "changes": {},
        "comment": "Group groupname is already present",
    }
    mocks["postgres.role_get"].assert_called_once()
    mocks["postgres.group_create"].assert_not_called()
    mocks["postgres.group_update"].assert_not_called()


def test_present_update_md5_password(mocks, existing_group, md5_pw, db_args):
    existing_group["password"] = "md500000000000000000000000000000000"
    mocks["postgres.role_get"].return_value = existing_group

    assert postgres_group.present("groupname", password="password", encrypted=True) == {
        "name": "groupname",
        "result": True,
        "changes": {"groupname": {"password": True}},
        "comment": "The group groupname has been updated",
    }
    mocks["postgres.role_get"].assert_called_once()
    mocks["postgres.group_create"].assert_not_called()
    mocks["postgres.group_update"].assert_called_once_with(
        groupname="groupname",
        createdb=None,
        createroles=None,
        encrypted=True,
        superuser=None,
        login=None,
        inherit=None,
        replication=None,
        rolepassword=md5_pw,
        groups=None,
        **db_args
    )


@pytest.mark.skip_on_fips_enabled_platform
def test_present_update_error(mocks, existing_group):
    existing_group["password"] = "md500000000000000000000000000000000"
    mocks["postgres.role_get"].return_value = existing_group
    mocks["postgres.group_update"].return_value = False

    assert postgres_group.present("groupname", password="password", encrypted=True) == {
        "name": "groupname",
        "result": False,
        "changes": {},
        "comment": "Failed to update group groupname",
    }
    mocks["postgres.role_get"].assert_called_once()
    mocks["postgres.group_create"].assert_not_called()
    mocks["postgres.group_update"].assert_called_once()


def test_present_update_password_no_check(mocks, existing_group, md5_pw, db_args):
    mocks["postgres.role_get"].return_value = existing_group

    assert postgres_group.present(
        "groupname", password="password", encrypted=True, refresh_password=True
    ) == {
        "name": "groupname",
        "result": True,
        "changes": {"groupname": {"password": True}},
        "comment": "The group groupname has been updated",
    }
    mocks["postgres.role_get"].assert_called_once_with(
        "groupname", return_password=False, **db_args
    )
    mocks["postgres.group_create"].assert_not_called()
    mocks["postgres.group_update"].assert_called_once_with(
        groupname="groupname",
        createdb=None,
        createroles=None,
        encrypted=True,
        superuser=None,
        login=None,
        inherit=None,
        replication=None,
        rolepassword=md5_pw,
        groups=None,
        **db_args
    )


# ==========
# postgres_group.absent
# ==========


def test_absent_delete(mocks, db_args):
    mocks["postgres.user_exists"].return_value = True

    assert postgres_group.absent("groupname") == {
        "name": "groupname",
        "result": True,
        "changes": {"groupname": "Absent"},
        "comment": "Group groupname has been removed",
    }
    mocks["postgres.user_exists"].assert_called_once_with("groupname", **db_args)
    mocks["postgres.group_remove"].assert_called_once_with("groupname", **db_args)


@pytest.mark.usefixtures("test_mode")
def test_absent_test(mocks, db_args):
    mocks["postgres.user_exists"].return_value = True

    assert postgres_group.absent("groupname") == {
        "name": "groupname",
        "result": None,
        "changes": {},
        "comment": "Group groupname is set to be removed",
    }
    mocks["postgres.user_exists"].assert_called_once_with("groupname", **db_args)
    mocks["postgres.group_remove"].assert_not_called()


def test_absent_already(mocks, db_args):
    mocks["postgres.user_exists"].return_value = False

    assert postgres_group.absent("groupname") == {
        "name": "groupname",
        "result": True,
        "changes": {},
        "comment": "Group groupname is not present, so it cannot be removed",
    }
    mocks["postgres.user_exists"].assert_called_once_with("groupname", **db_args)
    mocks["postgres.group_remove"].assert_not_called()


def test_absent_error(mocks):
    mocks["postgres.user_exists"].return_value = True
    mocks["postgres.group_remove"].return_value = False

    assert postgres_group.absent("groupname") == {
        "name": "groupname",
        "result": False,
        "changes": {},
        "comment": "Group groupname failed to be removed",
    }
    mocks["postgres.user_exists"].assert_called_once()
    mocks["postgres.group_remove"].assert_called_once()



"""
:codeauthor: Gareth J. Greenaway (ggreenaway@vmware.com)
"""

import logging
import random

import pytest
from pytestshellutils.exceptions import FactoryNotStarted
from saltfactories.utils import random_string

import salt.defaults.exitcodes
from tests.conftest import FIPS_TESTRUN
from tests.support.helpers import PRE_PYTEST_SKIP_REASON

log = logging.getLogger(__name__)


pytestmark = [
    pytest.mark.skip_on_spawning_platform(
        reason="Deltaproxy minions do not currently work on spawning platforms.",
    ),
    pytest.mark.core_test,
    pytest.mark.timeout_unless_on_windows(400),
]


@pytest.fixture(scope="package", autouse=True)
def _skip_on_tcp_transport(request):
    if request.config.getoption("--transport") == "tcp":
        pytest.skip("Deltaproxy under the TPC transport is not working. See #61367")


@pytest.fixture
def proxy_minion_id(salt_master):
    _proxy_minion_id = random_string("proxy-minion-")

    try:
        yield _proxy_minion_id
    finally:
        # Remove stale key if it exists
        pytest.helpers.remove_stale_minion_key(salt_master, _proxy_minion_id)


def test_exit_status_no_proxyid(salt_master, proxy_minion_id):
    """
    Ensure correct exit status when --proxyid argument is missing.
    """
    config_defaults = {
        "metaproxy": "deltaproxy",
    }

    with pytest.raises(FactoryNotStarted) as exc:
        factory = salt_master.salt_proxy_minion_daemon(
            proxy_minion_id, include_proxyid_cli_flag=False, defaults=config_defaults
        )
        factory.start(start_timeout=10, max_start_attempts=1)

    assert exc.value.process_result.returncode == salt.defaults.exitcodes.EX_USAGE
    assert "Usage" in exc.value.process_result.stderr, exc.value
    assert "error: salt-proxy requires --proxyid" in exc.value.process_result.stderr


@pytest.mark.skip_on_windows(reason="Windows does not do user checks")
def test_exit_status_unknown_user(salt_master, proxy_minion_id):
    """
    Ensure correct exit status when the proxy is configured to run as an
    unknown user.
    """
    config_defaults = {
        "metaproxy": "deltaproxy",
    }

    with pytest.raises(FactoryNotStarted) as exc:
        factory = salt_master.salt_proxy_minion_daemon(
            proxy_minion_id,
            overrides={"user": "unknown-user"},
            defaults=config_defaults,
        )
        factory.start(start_timeout=10, max_start_attempts=1)

    assert exc.value.process_result.returncode == salt.defaults.exitcodes.EX_NOUSER
    assert "The user is not available." in exc.value.process_result.stderr


def test_exit_status_unknown_argument(salt_master, proxy_minion_id):
    """
    Ensure correct exit status when an unknown argument is passed to
    salt-proxy.
    """
    config_defaults = {
        "metaproxy": "deltaproxy",
    }

    with pytest.raises(FactoryNotStarted) as exc:
        factory = salt_master.salt_proxy_minion_daemon(
            proxy_minion_id, defaults=config_defaults
        )
        factory.start("--unknown-argument", start_timeout=10, max_start_attempts=1)

    assert exc.value.process_result.returncode == salt.defaults.exitcodes.EX_USAGE
    assert "Usage" in exc.value.process_result.stderr
    assert "no such option: --unknown-argument" in exc.value.process_result.stderr


# Hangs on Windows. You can add a timeout to the proxy.run command, but then
# it just times out.
@pytest.mark.skip_on_windows(reason=PRE_PYTEST_SKIP_REASON)
@pytest.mark.parametrize(
    "parallel_startup",
    [True, False],
    ids=["parallel_startup=True", "parallel_startup=False"],
)
def test_exit_status_correct_usage(
    salt_master,
    salt_cli,
    proxy_minion_id,
    parallel_startup,
):
    """
    Ensure the salt-proxy control proxy starts and
    is able to respond to test.ping, additionally ensure that
    the proxies being controlled also respond to test.ping.

    Finally ensure correct exit status when salt-proxy exits correctly.

    Skip on Windows because daemonization not supported
    """

    config_defaults = {
        "metaproxy": "deltaproxy",
    }
    config_overrides = {
        "fips_mode": FIPS_TESTRUN,
        "encryption_algorithm": "OAEP-SHA224" if FIPS_TESTRUN else "OAEP-SHA1",
        "signing_algorithm": "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1",
    }
    proxy_one = "dummy_proxy_one"
    proxy_two = "dummy_proxy_two"

    top_file = """
    base:
      {control}:
        - controlproxy
      {one}:
        - {one}
      {two}:
        - {two}
    """.format(
        control=proxy_minion_id,
        one=proxy_one,
        two=proxy_two,
    )
    controlproxy_pillar_file = """
    proxy:
        proxytype: deltaproxy
        parallel_startup: {}
        ids:
          - {}
          - {}
    """.format(
        parallel_startup, proxy_one, proxy_two
    )

    dummy_proxy_one_pillar_file = """
    proxy:
      proxytype: dummy
    """

    dummy_proxy_two_pillar_file = """
    proxy:
      proxytype: dummy
    """

    top_tempfile = salt_master.pillar_tree.base.temp_file("top.sls", top_file)
    controlproxy_tempfile = salt_master.pillar_tree.base.temp_file(
        "controlproxy.sls", controlproxy_pillar_file
    )
    dummy_proxy_one_tempfile = salt_master.pillar_tree.base.temp_file(
        f"{proxy_one}.sls",
        dummy_proxy_one_pillar_file,
    )
    dummy_proxy_two_tempfile = salt_master.pillar_tree.base.temp_file(
        f"{proxy_two}.sls",
        dummy_proxy_two_pillar_file,
    )
    with (
        top_tempfile
    ), controlproxy_tempfile, dummy_proxy_one_tempfile, dummy_proxy_two_tempfile:
        factory = salt_master.salt_proxy_minion_daemon(
            proxy_minion_id,
            defaults=config_defaults,
            overrides=config_overrides,
            extra_cli_arguments_after_first_start_failure=["--log-level=info"],
            start_timeout=320,
        )

        for minion_id in (proxy_minion_id, proxy_one, proxy_two):
            factory.before_start(
                pytest.helpers.remove_stale_proxy_minion_cache_file, factory, minion_id
            )
            factory.after_terminate(
                pytest.helpers.remove_stale_minion_key, salt_master, minion_id
            )
            factory.after_terminate(
                pytest.helpers.remove_stale_proxy_minion_cache_file, factory, minion_id
            )

        with factory.started():
            assert factory.is_running()

            # Let's issue a ping the control proxy
            ret = salt_cli.run("test.ping", minion_tgt=proxy_minion_id)
            assert ret.returncode == 0
            assert ret.data is True

            # Let's issue a ping to one of the controlled proxies
            ret = salt_cli.run("test.ping", minion_tgt=proxy_one)
            assert ret.returncode == 0
            assert ret.data is True

            # Let's issue a ping to one of the controlled proxies
            ret = salt_cli.run("test.ping", minion_tgt=proxy_two)
            assert ret.returncode == 0
            assert ret.data is True

        # Terminate the proxy minion
        ret = factory.terminate()
        assert ret.returncode == salt.defaults.exitcodes.EX_OK, ret


# Hangs on Windows. You can add a timeout to the proxy.run command, but then
# it just times out.
@pytest.mark.skip_on_windows(reason=PRE_PYTEST_SKIP_REASON)
@pytest.mark.parametrize(
    "parallel_startup",
    [True, False],
    ids=["parallel_startup=True", "parallel_startup=False"],
)
def test_missing_pillar_file(
    salt_master,
    salt_cli,
    proxy_minion_id,
    parallel_startup,
):
    """
    Ensure that the control proxy minion starts up when
    pillar files for sub proxies are missing.

    Skip on Windows because daemonization not supported
    """

    config_defaults = {
        "metaproxy": "deltaproxy",
    }
    config_overrides = {
        "fips_mode": FIPS_TESTRUN,
        "encryption_algorithm": "OAEP-SHA224" if FIPS_TESTRUN else "OAEP-SHA1",
        "signing_algorithm": "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1",
    }
    proxy_one = "dummy_proxy_one"
    proxy_two = "dummy_proxy_two"

    top_file = """
    base:
      {control}:
        - controlproxy
      {one}:
        - {one}
    """.format(
        control=proxy_minion_id,
        one=proxy_one,
    )
    controlproxy_pillar_file = """
    proxy:
        proxytype: deltaproxy
        parallel_startup: {}
        ids:
          - {}
          - {}
    """.format(
        parallel_startup, proxy_one, proxy_two
    )

    dummy_proxy_one_pillar_file = """
    proxy:
      proxytype: dummy
    """

    top_tempfile = salt_master.pillar_tree.base.temp_file("top.sls", top_file)
    controlproxy_tempfile = salt_master.pillar_tree.base.temp_file(
        "controlproxy.sls", controlproxy_pillar_file
    )
    dummy_proxy_one_tempfile = salt_master.pillar_tree.base.temp_file(
        f"{proxy_one}.sls",
        dummy_proxy_one_pillar_file,
    )
    with top_tempfile, controlproxy_tempfile, dummy_proxy_one_tempfile:
        factory = salt_master.salt_proxy_minion_daemon(
            proxy_minion_id,
            defaults=config_defaults,
            overrides=config_overrides,
            extra_cli_arguments_after_first_start_failure=["--log-level=info"],
            start_timeout=320,
        )

        for minion_id in (proxy_minion_id, proxy_one, proxy_two):
            factory.before_start(
                pytest.helpers.remove_stale_proxy_minion_cache_file, factory, minion_id
            )
            factory.after_terminate(
                pytest.helpers.remove_stale_minion_key, salt_master, minion_id
            )
            factory.after_terminate(
                pytest.helpers.remove_stale_proxy_minion_cache_file, factory, minion_id
            )

        with factory.started():
            assert factory.is_running()

            # Let's issue a ping the control proxy
            ret = salt_cli.run("test.ping", minion_tgt=proxy_minion_id)
            assert ret.returncode == 0
            assert ret.data is True

            # Let's issue a ping to one of the controlled proxies
            ret = salt_cli.run("test.ping", minion_tgt="dummy_proxy_one")
            assert ret.returncode == 0
            assert ret.data is True

        # Terminate the proxy minion
        ret = factory.terminate()
        assert ret.returncode == salt.defaults.exitcodes.EX_OK, ret


# Hangs on Windows. You can add a timeout to the proxy.run command, but then
# it just times out.
@pytest.mark.skip_on_windows(reason=PRE_PYTEST_SKIP_REASON)
@pytest.mark.parametrize(
    "parallel_startup",
    [True, False],
    ids=["parallel_startup=True", "parallel_startup=False"],
)
def test_invalid_connection(
    salt_master,
    salt_cli,
    proxy_minion_id,
    parallel_startup,
):
    """
    Ensure that the control proxy minion starts up when
    pillar files for sub proxies are missing.

    Skip on Windows because daemonization not supported
    """

    config_defaults = {
        "metaproxy": "deltaproxy",
    }
    config_overrides = {
        "fips_mode": FIPS_TESTRUN,
        "encryption_algorithm": "OAEP-SHA224" if FIPS_TESTRUN else "OAEP-SHA1",
        "signing_algorithm": "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1",
    }
    proxy_one = "dummy_proxy_one"
    broken_proxy_one = "broken_proxy_one"
    broken_proxy_two = "broken_proxy_two"

    top_file = """
    base:
      {control}:
        - controlproxy
      {one}:
        - {one}
      {broken_proxy_one}:
        - {broken_proxy_one}
      {broken_proxy_two}:
        - {broken_proxy_two}
    """.format(
        control=proxy_minion_id,
        one=proxy_one,
        broken_proxy_one=broken_proxy_one,
        broken_proxy_two=broken_proxy_two,
    )
    controlproxy_pillar_file = """
    proxy:
        proxytype: deltaproxy
        parallel_startup: {}
        ids:
          - {}
          - {}
          - {}
    """.format(
        parallel_startup, broken_proxy_one, broken_proxy_two, proxy_one
    )

    dummy_proxy_one_pillar_file = """
    proxy:
      proxytype: dummy
    """

    broken_proxy_one_pillar_file = """
    proxy:
      proxytype: dummy
      raise_minion_error: True
    """

    broken_proxy_two_pillar_file = """
    proxy:
      proxytype: dummy
      raise_commandexec_error: True
    """

    top_tempfile = salt_master.pillar_tree.base.temp_file("top.sls", top_file)
    controlproxy_tempfile = salt_master.pillar_tree.base.temp_file(
        "controlproxy.sls", controlproxy_pillar_file
    )
    dummy_proxy_one_tempfile = salt_master.pillar_tree.base.temp_file(
        f"{proxy_one}.sls",
        dummy_proxy_one_pillar_file,
    )
    broken_proxy_one_tempfile = salt_master.pillar_tree.base.temp_file(
        f"{broken_proxy_one}.sls", broken_proxy_one_pillar_file
    )
    broken_proxy_two_tempfile = salt_master.pillar_tree.base.temp_file(
        f"{broken_proxy_two}.sls", broken_proxy_two_pillar_file
    )
    with (
        top_tempfile
    ), (
        controlproxy_tempfile
    ), dummy_proxy_one_tempfile, broken_proxy_one_tempfile, broken_proxy_two_tempfile:
        factory = salt_master.salt_proxy_minion_daemon(
            proxy_minion_id,
            defaults=config_defaults,
            overrides=config_overrides,
            extra_cli_arguments_after_first_start_failure=["--log-level=info"],
            start_timeout=320,
        )

        for minion_id in (
            proxy_minion_id,
            proxy_one,
            broken_proxy_one,
            broken_proxy_two,
        ):
            factory.before_start(
                pytest.helpers.remove_stale_proxy_minion_cache_file, factory, minion_id
            )
            factory.after_terminate(
                pytest.helpers.remove_stale_minion_key, salt_master, minion_id
            )
            factory.after_terminate(
                pytest.helpers.remove_stale_proxy_minion_cache_file, factory, minion_id
            )

        with factory.started():
            # Let's issue a ping the control proxy
            ret = salt_cli.run("test.ping", minion_tgt=proxy_minion_id)
            assert ret.returncode == 0
            assert ret.data is True
            # Let's issue a ping to one of the controlled proxies
            ret = salt_cli.run("test.ping", minion_tgt=proxy_one)
            assert ret.returncode == 0
            assert ret.data is True

    assert not factory.is_running()
    assert ret.returncode == salt.defaults.exitcodes.EX_OK, ret


@pytest.mark.skip_on_windows(reason=PRE_PYTEST_SKIP_REASON)
@pytest.mark.parametrize(
    "parallel_startup",
    [True, False],
    ids=["parallel_startup=True", "parallel_startup=False"],
)
def test_custom_proxy_module(
    salt_master,
    salt_cli,
    proxy_minion_id,
    parallel_startup,
):
    """
    Ensure the salt-proxy control proxy starts and
    is able to respond to test.ping, additionally ensure that
    the proxies being controlled also respond to test.ping.

    Finally ensure correct exit status when salt-proxy exits correctly.

    Skip on Windows because daemonization not supported
    """

    config_defaults = {
        "metaproxy": "deltaproxy",
    }
    config_overrides = {
        "fips_mode": FIPS_TESTRUN,
        "encryption_algorithm": "OAEP-SHA224" if FIPS_TESTRUN else "OAEP-SHA1",
        "signing_algorithm": "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1",
    }
    proxy_one = "custom_dummy_proxy_one"
    proxy_two = "custom_dummy_proxy_two"

    top_file = """
    base:
      {control}:
        - controlproxy
      {one}:
        - {one}
      {two}:
        - {two}
    """.format(
        control=proxy_minion_id,
        one=proxy_one,
        two=proxy_two,
    )
    controlproxy_pillar_file = """
    proxy:
        proxytype: deltaproxy
        parallel_startup: {}
        ids:
          - {}
          - {}
    """.format(
        parallel_startup, proxy_one, proxy_two
    )

    dummy_proxy_one_pillar_file = """
    proxy:
      proxytype: custom_dummy
    """

    dummy_proxy_two_pillar_file = """
    proxy:
      proxytype: custom_dummy
    """

    module_contents = """
__proxyenabled__ = ["custom_dummy"]

def __virtual__():
    return True

def init(opts):
    return True

def ping():
    return True
    """

    top_tempfile = salt_master.pillar_tree.base.temp_file("top.sls", top_file)
    controlproxy_tempfile = salt_master.pillar_tree.base.temp_file(
        "controlproxy.sls", controlproxy_pillar_file
    )
    dummy_proxy_one_tempfile = salt_master.pillar_tree.base.temp_file(
        f"{proxy_one}.sls",
        dummy_proxy_one_pillar_file,
    )
    dummy_proxy_two_tempfile = salt_master.pillar_tree.base.temp_file(
        f"{proxy_two}.sls",
        dummy_proxy_two_pillar_file,
    )

    custom_proxy_module = salt_master.state_tree.base.temp_file(
        "_proxy/custom_dummy.py", module_contents
    )
    with (
        top_tempfile
    ), (
        controlproxy_tempfile
    ), dummy_proxy_one_tempfile, dummy_proxy_two_tempfile, custom_proxy_module:
        factory = salt_master.salt_proxy_minion_daemon(
            proxy_minion_id,
            defaults=config_defaults,
            overrides=config_overrides,
            extra_cli_arguments_after_first_start_failure=["--log-level=info"],
            start_timeout=320,
        )

        for minion_id in (proxy_minion_id, proxy_one, proxy_two):
            factory.before_start(
                pytest.helpers.remove_stale_proxy_minion_cache_file, factory, minion_id
            )
            factory.after_terminate(
                pytest.helpers.remove_stale_minion_key, salt_master, minion_id
            )
            factory.after_terminate(
                pytest.helpers.remove_stale_proxy_minion_cache_file, factory, minion_id
            )

        with factory.started():
            assert factory.is_running()

            # Let's issue a ping the control proxy
            ret = salt_cli.run("test.ping", minion_tgt=proxy_minion_id)
            assert ret.returncode == 0
            assert ret.data is True

            # Let's issue a ping to one of the controlled proxies
            ret = salt_cli.run("test.ping", minion_tgt=proxy_one)
            assert ret.returncode == 0
            assert ret.data is True

            # Let's issue a ping to one of the controlled proxies
            ret = salt_cli.run("test.ping", minion_tgt=proxy_two)
            assert ret.returncode == 0
            assert ret.data is True

        # Terminate the proxy minion
        ret = factory.terminate()
        assert ret.returncode == salt.defaults.exitcodes.EX_OK, ret


@pytest.mark.skip_on_windows(reason=PRE_PYTEST_SKIP_REASON)
@pytest.mark.parametrize(
    "parallel_startup",
    [True, False],
    ids=["parallel_startup=True", "parallel_startup=False"],
)
def test_custom_proxy_module_raise_exception(
    salt_master,
    salt_cli,
    proxy_minion_id,
    parallel_startup,
):
    """
    Ensure the salt-proxy control proxy starts and
    is able to respond to test.ping, additionally ensure that
    the proxies being controlled also respond to test.ping.

    Finally ensure correct exit status when salt-proxy exits correctly.

    Skip on Windows because daemonization not supported
    """

    config_defaults = {
        "metaproxy": "deltaproxy",
    }
    config_overrides = {
        "fips_mode": FIPS_TESTRUN,
        "encryption_algorithm": "OAEP-SHA224" if FIPS_TESTRUN else "OAEP-SHA1",
        "signing_algorithm": "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1",
    }
    proxy_one = "custom_dummy_proxy_one"
    proxy_two = "custom_dummy_proxy_two"

    top_file = """
    base:
      {control}:
        - controlproxy
      {one}:
        - {one}
      {two}:
        - {two}
    """.format(
        control=proxy_minion_id,
        one=proxy_one,
        two=proxy_two,
    )
    controlproxy_pillar_file = """
    proxy:
        proxytype: deltaproxy
        parallel_startup: {}
        ids:
          - {}
          - {}
    """.format(
        parallel_startup, proxy_one, proxy_two
    )

    dummy_proxy_one_pillar_file = """
    proxy:
      proxytype: custom_dummy
    """

    dummy_proxy_two_pillar_file = """
    proxy:
      proxytype: dummy
    """

    module_contents = """
__proxyenabled__ = ["custom_dummy"]

def __virtual__():
    return True

def init(opts):
    raise Exception("Something has gone horribly wrong.")

def ping():
    return True
    """

    top_tempfile = salt_master.pillar_tree.base.temp_file("top.sls", top_file)
    controlproxy_tempfile = salt_master.pillar_tree.base.temp_file(
        "controlproxy.sls", controlproxy_pillar_file
    )
    dummy_proxy_one_tempfile = salt_master.pillar_tree.base.temp_file(
        f"{proxy_one}.sls",
        dummy_proxy_one_pillar_file,
    )
    dummy_proxy_two_tempfile = salt_master.pillar_tree.base.temp_file(
        f"{proxy_two}.sls",
        dummy_proxy_two_pillar_file,
    )

    custom_proxy_module = salt_master.state_tree.base.temp_file(
        "_proxy/custom_dummy.py", module_contents
    )
    with (
        top_tempfile
    ), (
        controlproxy_tempfile
    ), dummy_proxy_one_tempfile, dummy_proxy_two_tempfile, custom_proxy_module:
        factory = salt_master.salt_proxy_minion_daemon(
            proxy_minion_id,
            defaults=config_defaults,
            overrides=config_overrides,
            extra_cli_arguments_after_first_start_failure=["--log-level=info"],
            start_timeout=320,
        )

        for minion_id in (proxy_minion_id, proxy_one, proxy_two):
            factory.before_start(
                pytest.helpers.remove_stale_proxy_minion_cache_file, factory, minion_id
            )
            factory.after_terminate(
                pytest.helpers.remove_stale_minion_key, salt_master, minion_id
            )
            factory.after_terminate(
                pytest.helpers.remove_stale_proxy_minion_cache_file, factory, minion_id
            )

        with factory.started():
            assert factory.is_running()

            # Let's issue a ping the control proxy
            ret = salt_cli.run("test.ping", minion_tgt=proxy_minion_id)
            assert ret.returncode == 0
            assert ret.data is True

            # Let's issue a ping to one of the controlled proxies
            ret = salt_cli.run("test.ping", minion_tgt=proxy_one)
            assert ret.returncode == 1
            assert "Minion did not return" in ret.data

            # Let's issue a ping to one of the controlled proxies
            ret = salt_cli.run("test.ping", minion_tgt=proxy_two)
            assert ret.returncode == 0
            assert ret.data is True

        # Terminate the proxy minion
        ret = factory.terminate()
        assert ret.returncode == salt.defaults.exitcodes.EX_OK, ret


# Hangs on Windows. You can add a timeout to the proxy.run command, but then
# it just times out.
@pytest.mark.skip_on_windows(reason=PRE_PYTEST_SKIP_REASON)
@pytest.mark.parametrize(
    "parallel_startup",
    [True, False],
    ids=["parallel_startup=True", "parallel_startup=False"],
)
def test_exit_status_correct_usage_large_number_of_minions(
    salt_master,
    salt_cli,
    proxy_minion_id,
    parallel_startup,
):
    """
    Ensure the salt-proxy control proxy starts and
    is able to respond to test.ping, additionally ensure that
    the proxies being controlled also respond to test.ping.

    Finally ensure correct exit status when salt-proxy exits correctly.

    Skip on Windows because daemonization not supported
    """

    config_defaults = {
        "metaproxy": "deltaproxy",
    }
    config_overrides = {
        "fips_mode": FIPS_TESTRUN,
        "encryption_algorithm": "OAEP-SHA224" if FIPS_TESTRUN else "OAEP-SHA1",
        "signing_algorithm": "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1",
    }
    sub_proxies = [
        "proxy_one",
        "proxy_two",
        "proxy_three",
        "proxy_four",
        "proxy_five",
        "proxy_six",
        "proxy_seven",
        "proxy_eight",
        "proxy_nine",
        "proxy_ten",
        "proxy_eleven",
        "proxy_twelve",
        "proxy_thirteen",
        "proxy_fourteen",
        "proxy_fifteen",
        "proxy_sixteen",
        "proxy_seventeen",
        "proxy_eighteen",
        "proxy_nineteen",
        "proxy_twenty",
        "proxy_twenty_one",
        "proxy_twenty_two",
        "proxy_twenty_three",
        "proxy_twenty_four",
        "proxy_twenty_five",
        "proxy_twenty_six",
        "proxy_twenty_seven",
        "proxy_twenty_eight",
        "proxy_twenty_nine",
        "proxy_thirty",
        "proxy_thirty_one",
        "proxy_thirty_two",
    ]

    top_file = """
    base:
      {control}:
        - controlproxy
    """.format(
        control=proxy_minion_id,
    )
    controlproxy_pillar_file = f"""
    proxy:
        proxytype: deltaproxy
        parallel_startup: {parallel_startup}
        ids:
    """

    dummy_proxy_pillar_file = """
    proxy:
      proxytype: dummy
    """

    for minion_id in sub_proxies:
        top_file += f"""
      {minion_id}:
        - dummy"""

        controlproxy_pillar_file += f"""
            - {minion_id}
        """

    top_tempfile = salt_master.pillar_tree.base.temp_file("top.sls", top_file)
    controlproxy_tempfile = salt_master.pillar_tree.base.temp_file(
        "controlproxy.sls", controlproxy_pillar_file
    )
    dummy_proxy_tempfile = salt_master.pillar_tree.base.temp_file(
        "dummy.sls",
        dummy_proxy_pillar_file,
    )
    with top_tempfile, controlproxy_tempfile, dummy_proxy_tempfile:

        factory = salt_master.salt_proxy_minion_daemon(
            proxy_minion_id,
            defaults=config_defaults,
            overrides=config_overrides,
            extra_cli_arguments_after_first_start_failure=["--log-level=info"],
            start_timeout=320,
        )

        for minion_id in [proxy_minion_id] + sub_proxies:
            factory.before_start(
                pytest.helpers.remove_stale_proxy_minion_cache_file,
                factory,
                minion_id,
            )
            factory.after_terminate(
                pytest.helpers.remove_stale_minion_key, salt_master, minion_id
            )
            factory.after_terminate(
                pytest.helpers.remove_stale_proxy_minion_cache_file,
                factory,
                minion_id,
            )

        with factory.started():
            assert factory.is_running()

            # Let's issue a ping the control proxy
            ret = salt_cli.run("test.ping", minion_tgt=proxy_minion_id)
            assert ret.returncode == 0
            assert ret.data is True

            for minion_id in random.sample(sub_proxies, 4):
                # Let's issue a ping to one of the controlled proxies
                ret = salt_cli.run("test.ping", minion_tgt=minion_id)
                assert ret.returncode == 0
                assert ret.data is True

    # Terminate the proxy minion
    ret = factory.terminate()
    assert ret.returncode == salt.defaults.exitcodes.EX_OK, ret



"""
    Test cases for salt.returners.etcd_return

    :codeauthor: Caleb Beard <calebb@vmware.com>
"""

import copy

import pytest

import salt.returners.etcd_return as etcd_return
import salt.utils.etcd_util as etcd_util
import salt.utils.jid
import salt.utils.json
from tests.support.mock import MagicMock, call, create_autospec, patch


@pytest.fixture
def instance():
    return create_autospec(etcd_util.EtcdBase)


@pytest.fixture
def etcd_client_mock(instance):
    mocked_client = MagicMock()
    mocked_client.return_value = instance
    return mocked_client


@pytest.fixture
def profile_name():
    return "etcd_returner_profile"


@pytest.fixture
def returner_root():
    return "/salt/test-return"


@pytest.fixture
def etcd_config(profile_name, returner_root):
    return {
        profile_name: {
            "etcd.host": "127.0.0.1",
            "etcd.port": 2379,
        },
        "etcd.returner": profile_name,
        "etcd.returner_root": returner_root,
    }


@pytest.fixture
def configure_loader_modules():
    return {
        etcd_return: {
            "__opts__": {},
        },
    }


def test__get_conn(etcd_client_mock, profile_name, returner_root, instance):
    """
    Test the _get_conn utility function in etcd_return
    """
    with patch("salt.utils.etcd_util.get_conn", etcd_client_mock):
        # Test to make sure we get the right path back
        config = {
            profile_name: {"etcd.host": "127.0.0.1", "etcd.port": 2379},
            "etcd.returner": profile_name,
            "etcd.returner_root": returner_root,
        }
        assert etcd_return._get_conn(config, profile=profile_name) == (
            instance,
            returner_root,
        )

        # Test to make sure we get the default path back if none in opts
        config = {
            profile_name: {"etcd.host": "127.0.0.1", "etcd.port": 2379},
            "etcd.returner": profile_name,
        }
        assert etcd_return._get_conn(config, profile=profile_name) == (
            instance,
            "/salt/return",
        )


def test_returner(etcd_client_mock, instance, returner_root, profile_name, etcd_config):
    """
    Test the returner function in etcd_return
    """
    with patch("salt.utils.etcd_util.get_conn", etcd_client_mock):
        ret = {
            "id": "test-id",
            "jid": "123456789",
            "single-key": "single-value",
            "dict-key": {
                "dict-subkey-1": "subvalue-1",
                "dict-subkey-2": "subvalue-2",
            },
        }

        # Test returner with ttl in etcd config
        config = copy.deepcopy(etcd_config)
        config[profile_name]["etcd.ttl"] = 5
        config["etcd.returner_write_profile"] = profile_name

        with patch.dict(etcd_return.__opts__, config):
            assert etcd_return.returner(ret) is None
            dest = "/".join((returner_root, "jobs", ret["jid"], ret["id"], "{}"))
            calls = [
                call("/".join((returner_root, "minions", ret["id"])), ret["jid"], ttl=5)
            ] + [
                call(dest.format(key), salt.utils.json.dumps(ret[key]), ttl=5)
                for key in ret
            ]
            instance.set.assert_has_calls(calls, any_order=True)

        # Test returner with ttl in top level config
        config = copy.deepcopy(etcd_config)
        config["etcd.ttl"] = 6
        instance.set.reset_mock()
        with patch.dict(etcd_return.__opts__, config):
            assert etcd_return.returner(ret) is None
            dest = "/".join((returner_root, "jobs", ret["jid"], ret["id"], "{}"))
            calls = [
                call("/".join((returner_root, "minions", ret["id"])), ret["jid"], ttl=6)
            ] + [
                call(dest.format(key), salt.utils.json.dumps(ret[key]), ttl=6)
                for key in ret
            ]
            instance.set.assert_has_calls(calls, any_order=True)


def test_save_load(
    etcd_client_mock, instance, returner_root, profile_name, etcd_config
):
    """
    Test the save_load function in etcd_return
    """
    load = {
        "single-key": "single-value",
        "dict-key": {
            "dict-subkey-1": "subvalue-1",
            "dict-subkey-2": "subvalue-2",
        },
    }
    jid = "23"

    with patch("salt.utils.etcd_util.get_conn", etcd_client_mock):
        # Test save_load with ttl in etcd config
        config = copy.deepcopy(etcd_config)
        config[profile_name]["etcd.ttl"] = 5
        config["etcd.returner_write_profile"] = profile_name

        with patch.dict(etcd_return.__opts__, config):
            assert etcd_return.save_load(jid, load) is None
            instance.set.assert_called_with(
                "/".join((returner_root, "jobs", jid, ".load.p")),
                salt.utils.json.dumps(load),
                ttl=5,
            )

        # Test save_load with ttl in top level config
        config = copy.deepcopy(etcd_config)
        config["etcd.ttl"] = 6
        with patch.dict(etcd_return.__opts__, config):
            assert etcd_return.save_load(jid, load) is None
            instance.set.assert_called_with(
                "/".join((returner_root, "jobs", jid, ".load.p")),
                salt.utils.json.dumps(load),
                ttl=6,
            )

            # Test save_load with minion kwarg, unused at the moment
            assert (
                etcd_return.save_load(jid, load, minions=("minion-1", "minion-2"))
                is None
            )
            instance.set.assert_called_with(
                "/".join((returner_root, "jobs", jid, ".load.p")),
                salt.utils.json.dumps(load),
                ttl=6,
            )


def test_get_load(etcd_client_mock, instance, returner_root, profile_name, etcd_config):
    """
    Test the get_load function in etcd_return
    """
    load = {
        "single-key": "single-value",
        "dict-key": {
            "dict-subkey-1": "subvalue-1",
            "dict-subkey-2": "subvalue-2",
        },
    }
    instance.get.return_value = salt.utils.json.dumps(load)
    jid = "23"

    with patch("salt.utils.etcd_util.get_conn", etcd_client_mock):
        # Test get_load using etcd config
        config = copy.deepcopy(etcd_config)
        config["etcd.returner_read_profile"] = profile_name

        with patch.dict(etcd_return.__opts__, config):
            assert etcd_return.get_load(jid) == load
            instance.get.assert_called_with(
                "/".join((returner_root, "jobs", jid, ".load.p"))
            )

        # Test get_load using top level config profile name
        config = copy.deepcopy(etcd_config)
        with patch.dict(etcd_return.__opts__, config):
            assert etcd_return.get_load(jid) == load
            instance.get.assert_called_with(
                "/".join((returner_root, "jobs", jid, ".load.p"))
            )


def test_get_jid(etcd_client_mock, instance, returner_root, etcd_config):
    """
    Test the get_load function in etcd_return
    """
    jid = "10"

    with patch("salt.utils.etcd_util.get_conn", etcd_client_mock), patch.dict(
        etcd_return.__opts__, etcd_config
    ):
        # Test that no value for jid returns an empty dict
        with patch.object(instance, "get", return_value={}):
            assert etcd_return.get_jid(jid) == {}
            instance.get.assert_called_with(
                "/".join((returner_root, "jobs", jid)), recurse=True
            )

        # Test that a jid with child values returns them
        retval = {
            "test-id-1": {
                "return": salt.utils.json.dumps("test-return-1"),
            },
            "test-id-2": {
                "return": salt.utils.json.dumps("test-return-2"),
            },
        }

        with patch.object(instance, "get", return_value=retval):
            # assert etcd_return.get_jid(jid) == {}
            assert etcd_return.get_jid(jid) == {
                "test-id-1": {"return": "test-return-1"},
                "test-id-2": {"return": "test-return-2"},
            }
            instance.get.assert_called_with(
                "/".join((returner_root, "jobs", jid)), recurse=True
            )


def test_get_fun(etcd_client_mock, instance, returner_root, etcd_config):
    """
    Test the get_fun function in etcd_return
    """
    fun = "test.ping"

    with patch("salt.utils.etcd_util.get_conn", etcd_client_mock), patch.dict(
        etcd_return.__opts__, etcd_config
    ):
        # Test that no value for jid returns an empty dict
        with patch.object(instance, "get", return_value={}):
            assert etcd_return.get_fun(fun) == {}
            instance.get.assert_called_with(
                "/".join((returner_root, "minions")), recurse=True
            )

        # Test that a jid with child values returns them
        side_effect = (
            {
                "id-1": "1",
                "id-2": "2",
            },
            '"test.ping"',
            '"test.collatz"',
        )
        instance.get.reset_mock()

        with patch.object(instance, "get", side_effect=side_effect):
            # Could be either one depending on if Python<3.6
            retval = etcd_return.get_fun(fun)
            assert retval in [{"id-1": "test.ping"}, {"id-2": "test.ping"}]
            calls = [
                call("/".join((returner_root, "minions")), recurse=True),
                call("/".join((returner_root, "jobs", "1", "id-1", "fun"))),
                call("/".join((returner_root, "jobs", "2", "id-2", "fun"))),
            ]
            instance.get.assert_has_calls(calls, any_order=True)


def test_get_jids(etcd_client_mock, instance, returner_root, etcd_config):
    """
    Test the get_jids function in etcd_return
    """
    with patch("salt.utils.etcd_util.get_conn", etcd_client_mock), patch.dict(
        etcd_return.__opts__, etcd_config
    ):
        # Test that no value for jids returns an empty dict
        with patch.object(instance, "get", return_value={}):
            assert etcd_return.get_jids() == []
            instance.get.assert_called_with(
                "/".join((returner_root, "jobs")), recurse=True
            )

        # Test that having child job values returns them
        children = {
            "123": {},
            "456": "not a dictionary",
            "789": {},
        }

        with patch.object(instance, "get", return_value=children):
            retval = etcd_return.get_jids()
            assert len(retval) == 2
            assert "123" in retval
            assert "789" in retval
            instance.get.assert_called_with(
                "/".join((returner_root, "jobs")), recurse=True
            )


def test_get_minions(etcd_client_mock, instance, returner_root, etcd_config):
    """
    Test the get_minions function in etcd_return
    """
    with patch("salt.utils.etcd_util.get_conn", etcd_client_mock), patch.dict(
        etcd_return.__opts__, etcd_config
    ):
        # Test that no minions returns an empty dict
        with patch.object(instance, "get", return_value={}):
            assert etcd_return.get_minions() == []
            instance.get.assert_called_with(
                "/".join((returner_root, "minions")), recurse=True
            )

        # Test that having child minion values returns them
        children = {
            "id-1": "ignored-jid-1",
            "id-2": "ignored-jid-2",
        }
        with patch.object(instance, "get", return_value=children):
            retval = etcd_return.get_minions()
            assert len(retval) == 2
            assert "id-1" in retval
            assert "id-2" in retval
            instance.get.assert_called_with(
                "/".join((returner_root, "minions")), recurse=True
            )


def test_prep_jid():
    # Test that it returns a passed_jid if available
    assert etcd_return.prep_jid(passed_jid="23") == "23"

    # Test that giving `nocache` a value does nothing extra
    assert etcd_return.prep_jid(nocache=True, passed_jid="23") == "23"

    with patch.object(salt.utils.jid, "gen_jid", return_value="10"):
        assert etcd_return.prep_jid() == "10"



import pytest

from . import normalize_ret

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.core_test,
]


def test_requisites_onchanges_any(state, state_tree):
    """
    Call sls file containing several require_in and require.

    Ensure that some of them are failing and that the order is right.
    """
    sls_contents = """
    changing_state:
      cmd.run:
        - name: echo "Changed!"

    another_changing_state:
      cmd.run:
        - name: echo "Changed!"

    non_changing_state:
      test.succeed_without_changes:
        - comment: non_changing_state not changed

    another_non_changing_state:
      test.succeed_without_changes:
        - comment: another_non_changing_state not changed

    # Should succeed since at least one will have changes
    test_one_changing_states:
      cmd.run:
        - name: echo "Success!"
        - onchanges_any:
          - cmd: changing_state
          - cmd: another_changing_state
          - test: non_changing_state
          - test: another_non_changing_state

    test_two_non_changing_states:
      cmd.run:
        - name: echo "Should not run"
        - onchanges_any:
          - test: non_changing_state
          - test: another_non_changing_state
    """
    expected_result = {
        'cmd_|-another_changing_state_|-echo "Changed!"_|-run': {
            "__run_num__": 1,
            "changes": True,
            "comment": 'Command "echo "Changed!"" run',
            "result": True,
        },
        'cmd_|-changing_state_|-echo "Changed!"_|-run': {
            "__run_num__": 0,
            "changes": True,
            "comment": 'Command "echo "Changed!"" run',
            "result": True,
        },
        'cmd_|-test_one_changing_states_|-echo "Success!"_|-run': {
            "__run_num__": 4,
            "changes": True,
            "comment": 'Command "echo "Success!"" run',
            "result": True,
        },
        'cmd_|-test_two_non_changing_states_|-echo "Should not run"_|-run': {
            "__run_num__": 5,
            "changes": False,
            "comment": "State was not run because none of the onchanges reqs changed",
            "result": True,
        },
        "test_|-another_non_changing_state_|-another_non_changing_state_|-succeed_without_changes": {
            "__run_num__": 3,
            "changes": False,
            "comment": "another_non_changing_state not changed",
            "result": True,
        },
        "test_|-non_changing_state_|-non_changing_state_|-succeed_without_changes": {
            "__run_num__": 2,
            "changes": False,
            "comment": "non_changing_state not changed",
            "result": True,
        },
    }
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        result = normalize_ret(ret.raw)
        assert result == expected_result


def test_onchanges_requisite(state, state_tree):
    """
    Tests a simple state using the onchanges requisite
    """
    sls_contents = """
    changing_state:
      cmd.run:
        - name: echo "Changed!"

    non_changing_state:
      test.succeed_without_changes:
        - comment: non_changing_state not changed

    test_changing_state:
      cmd.run:
        - name: echo "Success!"
        - onchanges:
          - cmd: changing_state

    test_non_changing_state:
      cmd.run:
        - name: echo "Should not run"
        - onchanges:
          - test: non_changing_state
    """
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        assert (
            ret['cmd_|-test_changing_state_|-echo "Success!"_|-run'].comment
            == 'Command "echo "Success!"" run'
        )
        assert (
            ret['cmd_|-test_non_changing_state_|-echo "Should not run"_|-run'].comment
            == "State was not run because none of the onchanges reqs changed"
        )


def test_onchanges_requisite_multiple(state, state_tree):
    """
    Tests a simple state using the onchanges requisite
    """
    sls_contents = """
    changing_state:
      cmd.run:
        - name: echo "Changed!"

    another_changing_state:
      cmd.run:
        - name: echo "Changed!"

    non_changing_state:
      test.succeed_without_changes:
        - comment: non_changing_state not changed

    another_non_changing_state:
      test.succeed_without_changes:
        - comment: another_non_changing_state not changed

    test_two_changing_states:
      cmd.run:
        - name: echo "Success!"
        - onchanges:
          - cmd: changing_state
          - cmd: another_changing_state

    test_two_non_changing_states:
      cmd.run:
        - name: echo "Should not run"
        - onchanges:
          - test: non_changing_state
          - test: another_non_changing_state

    test_one_changing_state:
      cmd.run:
        - name: echo "Success!"
        - onchanges:
          - cmd: changing_state
          - test: non_changing_state
    """
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        assert (
            ret['cmd_|-test_two_changing_states_|-echo "Success!"_|-run'].comment
            == 'Command "echo "Success!"" run'
        )

        assert (
            ret[
                'cmd_|-test_two_non_changing_states_|-echo "Should not run"_|-run'
            ].comment
            == "State was not run because none of the onchanges reqs changed"
        )

        assert (
            ret['cmd_|-test_one_changing_state_|-echo "Success!"_|-run'].comment
            == 'Command "echo "Success!"" run'
        )


def test_onchanges_in_requisite(state, state_tree):
    """
    Tests a simple state using the onchanges_in requisite
    """
    sls_contents = """
    changing_state:
      cmd.run:
        - name: echo "Changed!"
        - onchanges_in:
          - cmd: test_changes_expected

    non_changing_state:
      test.succeed_without_changes:
        - comment: non_changing_state not changed
        - onchanges_in:
          - cmd: test_changes_not_expected

    test_changes_expected:
      cmd.run:
        - name: echo "Success!"

    test_changes_not_expected:
      cmd.run:
        - name: echo "Should not run"
    """
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        assert (
            ret['cmd_|-test_changes_expected_|-echo "Success!"_|-run'].comment
            == 'Command "echo "Success!"" run'
        )

        assert (
            ret['cmd_|-test_changes_not_expected_|-echo "Should not run"_|-run'].comment
            == "State was not run because none of the onchanges reqs changed"
        )


def test_onchanges_requisite_no_state_module(state, state_tree):
    """
    Tests a simple state using the onchanges requisite without state modules
    """
    sls_contents = """
    changing_state:
      cmd.run:
        - name: echo "Changed!"

    non_changing_state:
      test.succeed_without_changes:
        - comment: non_changing_state not changed

    test_changing_state:
      cmd.run:
        - name: echo "Success!"
        - onchanges:
          - changing_state

    test_non_changing_state:
      cmd.run:
        - name: echo "Should not run"
        - onchanges:
          - non_changing_state
    """
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        assert (
            ret['cmd_|-test_changing_state_|-echo "Success!"_|-run'].comment
            == 'Command "echo "Success!"" run'
        )


def test_onchanges_requisite_with_duration(state, state_tree):
    """
    Tests a simple state using the onchanges requisite
    the state will not run but results will include duration
    """
    sls_contents = """
    changing_state:
      cmd.run:
        - name: echo "Changed!"

    non_changing_state:
      test.succeed_without_changes:
        - comment: non_changing_state not changed

    test_changing_state:
      cmd.run:
        - name: echo "Success!"
        - onchanges:
          - cmd: changing_state

    test_non_changing_state:
      cmd.run:
        - name: echo "Should not run"
        - onchanges:
          - test: non_changing_state
    """
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        assert (
            "duration"
            in ret['cmd_|-test_non_changing_state_|-echo "Should not run"_|-run']
        )


def test_onchanges_any_recursive_error_issues_50811(state, state_tree):
    """
    test that onchanges_any does not causes a recursive error
    """
    sls_contents = """
    command-test:
      cmd.run:
        - name: ls
        - onchanges_any:
          - file: /tmp/an-unfollowed-file
    """
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
    assert ret["command-test"].result is False
